from typing import Type

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.micheline import MichelineLiteral
from pytezos.michelson.micheline import blind_unpack
from pytezos.michelson.micheline import parse_micheline_literal
from pytezos.michelson.micheline import parse_micheline_value
from pytezos.michelson.types.base import MichelsonType


class unit:
    def __repr__(self):
        return 'Unit'

    def __eq__(self, other):
        return isinstance(other, unit)

    def __lt__(self, other):
        return False


Unit = unit()


class TrueLiteral(Micheline, prim='True'):
    pass


class FalseLiteral(Micheline, prim='False'):
    pass


class UnitLiteral(Micheline, prim='Unit'):
    pass


class StringType(MichelsonType, prim='string'):
    def __init__(self, value: str = ''):
        super(StringType, self).__init__()
        self.value = value

    def __lt__(self, other: 'StringType'):  # type: ignore
        return self.value < other.value

    def __eq__(self, other):  # type: ignore
        if not isinstance(other, StringType):
            return False
        return self.value == other.value

    def __hash__(self):
        return hash(self.value)

    def __repr__(self):
        return f'\'{self.value}\''

    def __str__(self):
        return self.value

    def __len__(self):
        return len(self.value)

    @classmethod
    def from_value(cls, value: str) -> 'StringType':
        assert isinstance(value, str), f'expected string, got {type(value).__name__}'
        assert len(value) == len(value.encode()), f'unicode symbols are not allowed: {value}'
        return cls(value)

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'StringType':
        return cls()

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'StringType':
        value = parse_micheline_literal(val_expr, {'string': str})
        return cls.from_value(value)

    @classmethod
    def from_python_object(cls, py_obj) -> 'StringType':
        return cls.from_value(py_obj)

    def to_literal(self) -> Type[Micheline]:
        return MichelineLiteral.create(self.value)

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        return {'string': self.value}

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        return self.value

    def __getitem__(self, item):
        assert isinstance(item, slice), f'expected start:end, got {item}'
        assert len(self.value) > 0, f'string is empty'
        assert item.stop <= len(self.value), f'out of bounds {item.stop} <= {len(self.value)}'
        return StringType(self.value[item.start : item.stop])


class IntType(MichelsonType, prim='int'):
    def __init__(self, value: int = 0):
        super(IntType, self).__init__()
        self.value = value

    def __lt__(self, other: 'IntType'):  # type: ignore
        return self.value < other.value

    def __eq__(self, other):  # type: ignore
        if not isinstance(other, IntType):
            return False
        return self.value == other.value

    def __hash__(self):
        return hash(self.value)

    def __repr__(self):
        return str(self.value)

    def __int__(self):
        return self.value

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'IntType':
        return cls()

    @classmethod
    def from_value(cls, value: int) -> 'IntType':
        return cls(value)

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'IntType':
        value = parse_micheline_literal(val_expr, {'int': int})
        return cls(value)

    @classmethod
    def from_python_object(cls, py_obj) -> 'IntType':
        assert isinstance(py_obj, int), f'expected integer, got {type(py_obj).__name__}'
        return cls(py_obj)

    def to_literal(self) -> Type[Micheline]:
        return MichelineLiteral.create(self.value)

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        return {'int': str(self.value)}

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        return self.value


class NatType(IntType, prim='nat'):
    @classmethod
    def from_value(cls, value: int) -> 'NatType':
        assert value >= 0, f'expected natural number, got {value}'
        return cls(value)

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'NatType':
        value = parse_micheline_literal(val_expr, {'int': int})
        return cls.from_value(value)

    @classmethod
    def from_python_object(cls, py_obj) -> 'NatType':
        assert isinstance(py_obj, int), f'expected integer, got {type(py_obj).__name__}'
        return cls.from_value(py_obj)


class BytesType(MichelsonType, prim='bytes'):
    def __init__(self, value: bytes = b''):
        super(BytesType, self).__init__()
        self.value = value

    def __lt__(self, other: 'BytesType'):  # type: ignore
        return self.value < other.value

    def __eq__(self, other):  # type: ignore
        if not isinstance(other, BytesType):
            return False
        return self.value == other.value

    def __hash__(self):
        return hash(self.value)

    def __repr__(self):
        return f'0x{self.value.hex()}'

    def __bytes__(self):
        return self.value

    def __len__(self):
        return len(self.value)

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'BytesType':
        return cls()

    @classmethod
    def from_value(cls, value: bytes):
        return cls(value)

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'BytesType':
        value = parse_micheline_literal(val_expr, {'bytes': bytes.fromhex})
        return cls(value)

    @classmethod
    def from_python_object(cls, py_obj) -> 'BytesType':
        if isinstance(py_obj, bytes):
            value = py_obj
        elif isinstance(py_obj, str):
            if py_obj.startswith('0x'):
                py_obj = py_obj[2:]
            value = bytes.fromhex(py_obj)
        else:
            raise AssertionError(f'unexpected value type {py_obj}')
        return cls(value)

    def to_literal(self) -> Type[Micheline]:
        return MichelineLiteral.create(self.value)

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        return {'bytes': self.value.hex()}

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        if try_unpack:
            return blind_unpack(self.value)
        return self.value

    def __getitem__(self, item):
        assert isinstance(item, slice), f'expected start:stop, got {item}'
        assert item.stop <= len(self.value), f'index out of bounds'
        return BytesType(self.value[item.start : item.stop])


class BoolType(MichelsonType, prim='bool'):
    def __init__(self, value: bool):
        super(BoolType, self).__init__()
        self.value = value

    def __lt__(self, other: 'BoolType'):  # type: ignore
        return self.value < other.value

    def __eq__(self, other):  # type: ignore
        if not isinstance(other, BoolType):
            return False
        return self.value == other.value

    def __hash__(self):
        return hash(self.value)

    def __repr__(self):
        return str(self.value)

    def __bool__(self):
        return self.value

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'BoolType':
        return cls(False)

    @classmethod
    def from_value(cls, value: bool):
        return cls(value)

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'BoolType':
        value = parse_micheline_value(
            val_expr,
            {
                ('False', 0): lambda x: False,
                ('True', 0): lambda x: True,
            },
        )
        return cls(value)

    @classmethod
    def from_python_object(cls, py_obj) -> 'BoolType':
        assert isinstance(py_obj, bool), f'expected boolean, got {type(py_obj).__name__}'
        return cls(py_obj)

    def to_literal(self) -> Type[Micheline]:
        return TrueLiteral if self.value else FalseLiteral

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        return {'prim': 'True' if self.value else 'False'}

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        return self.value


class UnitType(MichelsonType, prim='unit'):
    def __init__(self):
        super(UnitType, self).__init__()

    def __lt__(self, other: 'UnitType'):  # type: ignore
        return False

    def __eq__(self, other: 'UnitType'):  # type: ignore
        return True

    def __hash__(self):
        return hash(Unit)

    def __repr__(self):
        return 'Unit'

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'UnitType':
        return cls()

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'UnitType':
        parse_micheline_value(val_expr, {('Unit', 0): lambda x: x})
        return cls()

    @classmethod
    def from_python_object(cls, py_obj) -> 'UnitType':
        assert py_obj is None or isinstance(py_obj, unit), f'expected None or Unit, got {type(py_obj).__name__}'
        return cls()

    def to_literal(self) -> Type[Micheline]:
        return UnitLiteral

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        return {'prim': 'Unit'}

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        return unit()


class NeverType(MichelsonType, prim='never'):
    def __lt__(self, other: 'NeverType'):  # type: ignore
        return False

    def __eq__(self, other: 'NeverType'):  # type: ignore
        return True



from copy import copy
from typing import Generator
from typing import List
from typing import Type

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.micheline import MichelineSequence
from pytezos.michelson.types.base import MichelsonType


class SetType(MichelsonType, prim='set', args_len=1):
    def __init__(self, items: List[MichelsonType]):
        super(SetType, self).__init__()
        self.items = items

    def __repr__(self):
        return f'{{{", ".join(map(repr, self.items))}}}'

    def __len__(self):
        return len(self.items)

    def __iter__(self) -> Generator[MichelsonType, None, None]:
        yield from iter(self.items)

    def __eq__(self, other) -> bool:
        if not isinstance(other, SetType):
            return False
        return self.items == other.items

    @staticmethod
    def empty(item_type: Type[MichelsonType]) -> 'SetType':
        cls = SetType.create_type(args=[item_type])
        return cls(items=[])  # type: ignore

    @staticmethod
    def from_items(items: List[MichelsonType]) -> 'SetType':
        assert len(items) > 0, 'cannot instantiate from empty list'
        item_type = items[0].get_anon_type()
        for item in items[1:]:
            item_type.assert_type_equal(type(item))
        cls = SetType.create_type(args=[item_type])
        cls.check_constraints(items)  # type: ignore
        return cls(items)  # type: ignore

    @classmethod
    def check_constraints(cls, items: List[MichelsonType]):
        assert len(set(items)) == len(items), f'duplicate elements found'
        assert items == sorted(items), f'set elements are not sorted'

    @classmethod
    def dummy(cls, context: AbstractContext):
        return cls([])

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'SetType':
        assert isinstance(val_expr, list), f'expected list, got {type(val_expr).__name__}'
        items = list(map(cls.args[0].from_micheline_value, val_expr))
        cls.check_constraints(items)
        return cls(items)

    @classmethod
    def from_python_object(cls, py_obj) -> 'SetType':
        if isinstance(py_obj, list):
            py_set = set(py_obj)
            assert len(py_set) == len(py_obj), f'found duplicates'
        else:
            assert isinstance(py_obj, set), f'expected set or list, got {type(py_obj).__name__}'
            py_set = py_obj
        items = list(map(cls.args[0].from_python_object, py_set))
        items = sorted(items)
        return cls(items)

    def to_literal(self) -> Type[Micheline]:
        return MichelineSequence.create_type(args=[item.to_literal() for item in self.items])

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        return list(map(lambda x: x.to_micheline_value(mode=mode, lazy_diff=lazy_diff), self))

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        assert not comparable, f'{self.prim} is not comparable'
        return list(
            map(
                lambda x: x.to_python_object(
                    try_unpack=try_unpack,
                    lazy_diff=lazy_diff,
                    comparable=True,
                ),
                self,
            )
        )

    @classmethod
    def generate_pydoc(cls, definitions: list, inferred_name=None, comparable=False):
        name = cls.field_name or cls.type_name or inferred_name
        arg_doc = cls.args[0].generate_pydoc(
            definitions, inferred_name=f'{name}_item' if name else None, comparable=True
        )
        return f'{{ {arg_doc}, … }}'

    def contains(self, item: MichelsonType) -> bool:
        self.args[0].assert_type_equal(type(item))
        return item in self.items

    def add(self, item: MichelsonType) -> 'SetType':
        if self.contains(item):
            return copy(self)
        else:
            items = [item] + self.items
            return type(self)(sorted(items))

    def remove(self, item: MichelsonType) -> 'SetType':
        if self.contains(item):
            items = list(filter(lambda x: x != item, self.items))
            return type(self)(items)
        else:
            return copy(self)

    def __contains__(self, py_obj):
        key = self.args[0].from_python_object(py_obj)
        return self.contains(key)



from pytezos.michelson.instructions.adt import CarInstruction
from pytezos.michelson.instructions.adt import CdrInstruction
from pytezos.michelson.instructions.adt import GetnInstruction
from pytezos.michelson.instructions.adt import LeftInstruction
from pytezos.michelson.instructions.adt import PairInstruction
from pytezos.michelson.instructions.adt import RightInstruction
from pytezos.michelson.instructions.adt import UnpairInstruction
from pytezos.michelson.instructions.adt import UpdatenInstruction
from pytezos.michelson.instructions.arithmetic import AbsInstruction
from pytezos.michelson.instructions.arithmetic import AddInstruction
from pytezos.michelson.instructions.arithmetic import EdivInstruction
from pytezos.michelson.instructions.arithmetic import IntInstruction
from pytezos.michelson.instructions.arithmetic import IsNatInstruction
from pytezos.michelson.instructions.arithmetic import LslInstruction
from pytezos.michelson.instructions.arithmetic import LsrInstruction
from pytezos.michelson.instructions.arithmetic import MulInstruction
from pytezos.michelson.instructions.arithmetic import NegInstruction
from pytezos.michelson.instructions.arithmetic import SubInstruction
from pytezos.michelson.instructions.boolean import AndInstruction
from pytezos.michelson.instructions.boolean import NotInstruction
from pytezos.michelson.instructions.boolean import OrInstruction
from pytezos.michelson.instructions.boolean import XorInstruction
from pytezos.michelson.instructions.compare import CompareInstruction
from pytezos.michelson.instructions.compare import EqInstruction
from pytezos.michelson.instructions.compare import GeInstruction
from pytezos.michelson.instructions.compare import GtInstruction
from pytezos.michelson.instructions.compare import LeInstruction
from pytezos.michelson.instructions.compare import LtInstruction
from pytezos.michelson.instructions.compare import NeqInstruction
from pytezos.michelson.instructions.control import ApplyInstruction
from pytezos.michelson.instructions.control import DipInstruction
from pytezos.michelson.instructions.control import DipnInstruction
from pytezos.michelson.instructions.control import ExecInstruction
from pytezos.michelson.instructions.control import FailwithInstruction
from pytezos.michelson.instructions.control import IfConsInstruction
from pytezos.michelson.instructions.control import IfInstruction
from pytezos.michelson.instructions.control import IfLeftInstruction
from pytezos.michelson.instructions.control import IfNoneInstruction
from pytezos.michelson.instructions.control import IterInstruction
from pytezos.michelson.instructions.control import LambdaInstruction
from pytezos.michelson.instructions.control import LoopInstruction
from pytezos.michelson.instructions.control import LoopLeftInstruction
from pytezos.michelson.instructions.control import MapInstruction
from pytezos.michelson.instructions.control import PushInstruction
from pytezos.michelson.instructions.crypto import Blake2bInstruction
from pytezos.michelson.instructions.crypto import CheckSignatureInstruction
from pytezos.michelson.instructions.crypto import HashKeyInstruction
from pytezos.michelson.instructions.crypto import KeccakInstruction
from pytezos.michelson.instructions.crypto import PairingCheckInstruction
from pytezos.michelson.instructions.crypto import SaplingEmptyStateInstruction
from pytezos.michelson.instructions.crypto import SaplingVerifyUpdateInstruction
from pytezos.michelson.instructions.crypto import Sha3Instruction
from pytezos.michelson.instructions.crypto import Sha256Instruction
from pytezos.michelson.instructions.crypto import Sha512Instruction
from pytezos.michelson.instructions.generic import ConcatInstruction
from pytezos.michelson.instructions.generic import NeverInstruction
from pytezos.michelson.instructions.generic import PackInstruction
from pytezos.michelson.instructions.generic import SizeInstruction
from pytezos.michelson.instructions.generic import SliceInstruction
from pytezos.michelson.instructions.generic import UnitInstruction
from pytezos.michelson.instructions.generic import UnpackInstruction
from pytezos.michelson.instructions.jupyter import BeginInstruction
from pytezos.michelson.instructions.jupyter import BigMapDiffInstruction
from pytezos.michelson.instructions.jupyter import CommitInstruction
from pytezos.michelson.instructions.jupyter import DebugInstruction
from pytezos.michelson.instructions.jupyter import DropAllInstruction
from pytezos.michelson.instructions.jupyter import DumpAllInstruction
from pytezos.michelson.instructions.jupyter import DumpInstruction
from pytezos.michelson.instructions.jupyter import PatchInstruction
from pytezos.michelson.instructions.jupyter import PatchValueInstruction
from pytezos.michelson.instructions.jupyter import PrintInstruction
from pytezos.michelson.instructions.jupyter import ResetInstruction
from pytezos.michelson.instructions.jupyter import ResetValueInstruction
from pytezos.michelson.instructions.jupyter import RunInstruction
from pytezos.michelson.instructions.stack import DigInstruction
from pytezos.michelson.instructions.stack import DropInstruction
from pytezos.michelson.instructions.stack import DropnInstruction
from pytezos.michelson.instructions.stack import DugInstruction
from pytezos.michelson.instructions.stack import DupInstruction
from pytezos.michelson.instructions.stack import DupnInstruction
from pytezos.michelson.instructions.stack import PushInstruction
from pytezos.michelson.instructions.stack import RenameInstruction
from pytezos.michelson.instructions.stack import SwapInstruction
from pytezos.michelson.instructions.struct import ConsInstruction
from pytezos.michelson.instructions.struct import EmptyBigMapInstruction
from pytezos.michelson.instructions.struct import EmptyMapInstruction
from pytezos.michelson.instructions.struct import EmptySetInstruction
from pytezos.michelson.instructions.struct import GetAndUpdateInstruction
from pytezos.michelson.instructions.struct import GetInstruction
from pytezos.michelson.instructions.struct import MemInstruction
from pytezos.michelson.instructions.struct import NilInstruction
from pytezos.michelson.instructions.struct import NoneInstruction
from pytezos.michelson.instructions.struct import SomeInstruction
from pytezos.michelson.instructions.struct import UpdateInstruction
from pytezos.michelson.instructions.tezos import AddressInstruction
from pytezos.michelson.instructions.tezos import AmountInstruction
from pytezos.michelson.instructions.tezos import BalanceInstruction
from pytezos.michelson.instructions.tezos import ChainIdInstruction
from pytezos.michelson.instructions.tezos import ContractInstruction
from pytezos.michelson.instructions.tezos import CreateContractInstruction
from pytezos.michelson.instructions.tezos import ImplicitAccountInstruction
from pytezos.michelson.instructions.tezos import MinBlockTimeInstruction
from pytezos.michelson.instructions.tezos import NowInstruction
from pytezos.michelson.instructions.tezos import SelfAddressInstruction
from pytezos.michelson.instructions.tezos import SelfInstruction
from pytezos.michelson.instructions.tezos import SenderInstruction
from pytezos.michelson.instructions.tezos import SetDelegateInstruction
from pytezos.michelson.instructions.tezos import SourceInstruction
from pytezos.michelson.instructions.tezos import TotalVotingPowerInstruction
from pytezos.michelson.instructions.tezos import TransferTokensInstruction
from pytezos.michelson.instructions.tezos import VotingPowerInstruction
from pytezos.michelson.instructions.ticket import JoinTicketsInstruction
from pytezos.michelson.instructions.ticket import ReadTicketInstruction
from pytezos.michelson.instructions.ticket import SplitTicketInstruction
from pytezos.michelson.instructions.ticket import TicketInstruction
from pytezos.michelson.instructions.tzt import BigMapInstruction
from pytezos.michelson.instructions.tzt import StackEltInstruction



from typing import List
from typing import Tuple
from typing import Union
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.base import dispatch_types
from pytezos.michelson.instructions.base import format_stdout
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types import BytesType
from pytezos.michelson.types import ListType
from pytezos.michelson.types import MapType
from pytezos.michelson.types import NatType
from pytezos.michelson.types import NeverType
from pytezos.michelson.types import OptionType
from pytezos.michelson.types import SetType
from pytezos.michelson.types import StringType
from pytezos.michelson.types import UnitType


class ConcatInstruction(MichelsonInstruction, prim='CONCAT'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a = cast(Union[StringType, BytesType, ListType], stack.pop1())
        a.assert_type_in(StringType, BytesType, ListType)
        if isinstance(a, ListType):
            a.assert_type_in(ListType)
            res_type, convert, delim = dispatch_types(
                a.args[0],
                mapping={
                    (StringType,): (StringType, str, ''),
                    (BytesType,): (BytesType, bytes, b''),
                },
            )
            res = res_type.from_value(delim.join(map(convert, a)))
            stdout.append(format_stdout(cls.prim, [a], [res]))  # type: ignore
        else:
            b = cast(Union[StringType, BytesType], stack.pop1())
            res_type, convert = dispatch_types(
                type(a),
                type(b),
                mapping={
                    (StringType, StringType): (StringType, str),
                    (BytesType, BytesType): (BytesType, bytes),
                },
            )
            res = res_type.from_value(convert(a) + convert(b))
            stdout.append(format_stdout(cls.prim, [a, b], [res]))  # type: ignore
        stack.push(res)
        return cls(stack_items_added=1)


class PackInstruction(MichelsonInstruction, prim='PACK'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a = stack.pop1()
        res = BytesType.from_value(a.pack())
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [a], [res]))  # type: ignore
        return cls(stack_items_added=1)


class UnpackInstruction(MichelsonInstruction, prim='UNPACK', args_len=1):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a = cast(BytesType, stack.pop1())
        a.assert_type_equal(BytesType)
        try:
            some = cls.args[0].unpack(bytes(a))  # type: ignore
            res = OptionType.from_some(some)
        except Exception as e:
            stdout.append(f'{cls.prim}: {e}')
            res = OptionType.none(cls.args[0])  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [a], [res]))  # type: ignore
        return cls(stack_items_added=1)


class SizeInstruction(MichelsonInstruction, prim='SIZE'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        src = cast(Union[StringType, BytesType, ListType, SetType, MapType], stack.pop1())
        src.assert_type_in(StringType, BytesType, ListType, SetType, MapType)
        res = NatType.from_value(len(src))
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [src], [res]))  # type: ignore
        return cls(stack_items_added=1)


class SliceInstruction(MichelsonInstruction, prim='SLICE'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        offset, length, s = cast(Tuple[NatType, NatType, Union[StringType, BytesType]], stack.pop3())
        offset.assert_type_equal(NatType)
        length.assert_type_equal(NatType)
        s.assert_type_in(StringType, BytesType)
        start, stop = int(offset), int(offset) + int(length)
        if 0 <= start <= stop <= len(s):
            res = OptionType.from_some(s[start:stop])
        else:
            res = OptionType.none(type(s))
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [offset, length, s], [res]))  # type: ignore
        return cls(stack_items_added=1)


class UnitInstruction(MichelsonInstruction, prim='UNIT'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        res = UnitType()
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class NeverInstruction(MichelsonInstruction, prim='NEVER'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        never = cast(NeverType, stack.pop1())
        never.assert_type_equal(NeverType)
        stdout.append(format_stdout(cls.prim, [never], []))  # type: ignore
        return cls()



from pytezos.michelson.types.core import BytesType


class ChestType(BytesType, prim='chest'):
    ...
    # TODO: https://gitlab.com/tezos/tezos/-/merge_requests/2940/diffs#2c09e5627158501e568f7f4a7c9245c90c357217


class ChestKeyType(BytesType, prim='chest_key'):
    ...
    # TODO:



from typing import Callable
from typing import List
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.base import format_stdout
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types import BoolType
from pytezos.michelson.types import IntType


def compare(a, b) -> int:
    if a == b:
        return 0
    elif a < b:
        return -1
    else:
        return 1


class CompareInstruction(MichelsonInstruction, prim='COMPARE'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a, b = stack.pop2()
        a.assert_type_equal(type(b))
        res = IntType.from_value(compare(a, b))
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [a, b], [res]))  # type: ignore
        return cls(stack_items_added=1)


def execute_zero_compare(prim: str, stack: MichelsonStack, stdout: List[str], compare: Callable[[int], bool]):
    a = cast(IntType, stack.pop1())
    a.assert_type_equal(IntType)
    res = BoolType(compare(int(a)))
    stack.push(res)
    stdout.append(format_stdout(prim, [a], [res]))


class EqInstruction(MichelsonInstruction, prim='EQ'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        execute_zero_compare(cls.prim, stack, stdout, lambda x: x == 0)  # type: ignore
        return cls(stack_items_added=1)


class GeInstruction(MichelsonInstruction, prim='GE'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        execute_zero_compare(cls.prim, stack, stdout, lambda x: x >= 0)  # type: ignore
        return cls(stack_items_added=1)


class GtInstruction(MichelsonInstruction, prim='GT'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        execute_zero_compare(cls.prim, stack, stdout, lambda x: x > 0)  # type: ignore
        return cls(stack_items_added=1)


class LeInstruction(MichelsonInstruction, prim='LE'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        execute_zero_compare(cls.prim, stack, stdout, lambda x: x <= 0)  # type: ignore
        return cls(stack_items_added=1)


class LtInstruction(MichelsonInstruction, prim='LT'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        execute_zero_compare(cls.prim, stack, stdout, lambda x: x < 0)  # type: ignore
        return cls(stack_items_added=1)


class NeqInstruction(MichelsonInstruction, prim='NEQ'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        execute_zero_compare(cls.prim, stack, stdout, lambda x: x != 0)  # type: ignore
        return cls(stack_items_added=1)



import json
import logging
from decimal import Decimal
from functools import cached_property
from functools import lru_cache
from os.path import exists
from os.path import expanduser
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union
from typing import cast
from urllib.parse import urlparse

import requests
from deprecation import deprecated  # type: ignore

from pytezos.context.mixin import ContextMixin
from pytezos.context.mixin import ExecutionContext
from pytezos.contract.data import ContractData
from pytezos.contract.entrypoint import ContractEntrypoint
from pytezos.contract.metadata import ContractMetadata
from pytezos.contract.result import ContractCallResult
from pytezos.contract.token_metadata import ContractTokenMetadata
from pytezos.contract.view import ContractView
from pytezos.crypto.key import Key
from pytezos.jupyter import get_class_docstring
from pytezos.logging import logger
from pytezos.michelson.format import micheline_to_michelson
from pytezos.michelson.micheline import MichelsonRuntimeError
from pytezos.michelson.parse import michelson_to_micheline
from pytezos.michelson.program import MichelsonProgram
from pytezos.michelson.sections import ViewSection
from pytezos.michelson.types import BigMapType
from pytezos.michelson.types import BytesType
from pytezos.michelson.types.base import generate_pydoc
from pytezos.operation.group import OperationGroup
from pytezos.rpc import ShellQuery


class ContractTokenMetadataProxy:
    """Get TZIP-21 contract token metadata by token_id"""

    def __init__(self, fn: Callable) -> None:
        self._fn = fn

    def __getitem__(self, item):
        return self._fn(item)


class ContractInterface(ContextMixin):
    """Proxy class for interacting with a contract."""

    program: MichelsonProgram

    def __init__(self, context: ExecutionContext) -> None:
        super().__init__(context=context)
        self._logger = logging.getLogger(__name__)
        self._storage: Optional[ContractData] = None
        self.entrypoints = self.program.parameter.list_entrypoints()
        self.views = {view.name: view for view in self.program.views}  # type: Dict[str, Type[ViewSection]]

        for entrypoint, ty in self.entrypoints.items():
            if entrypoint == 'token_metadata':
                continue
            attr = ContractEntrypoint(context=context, entrypoint=entrypoint)
            attr.__doc__ = generate_pydoc(ty, entrypoint)
            assert not hasattr(self, entrypoint), f'Entrypoint name collision {entrypoint}'
            setattr(self, entrypoint, attr)

        for view_name, view_ty in self.views.items():
            view_attr = ContractView(
                context=context,
                name=view_name,
                parameter=view_ty.args[1].as_micheline_expr(),
                return_type=view_ty.args[2].as_micheline_expr(),
                code=view_ty.args[3].as_micheline_expr(),  # type: ignore
            )
            view_attr.__doc__ = view_ty.generate_pydoc()  # type: ignore
            assert not hasattr(self, view_name), f'View name collision {view_name}'
            setattr(self, view_name, view_attr)

    def __repr__(self) -> str:
        res = [
            super().__repr__(),
            '.storage\t# access storage data at block `block_id`',
            '.parameter\t# root entrypoint',
            '\nEntrypoints',
            *list(map(lambda x: f'.{x}()', self.entrypoints)),
            '\nViews',
            *list(map(lambda x: f'.{x}()', self.views)),
            '\nHelpers',
            get_class_docstring(self.__class__, attr_filter=lambda x: x not in self.entrypoints),
        ]
        return '\n'.join(res)

    def __getattr__(self, item: str) -> ContractEntrypoint:
        raise AttributeError(f'unexpected entrypoint {item}')

    @staticmethod
    def from_url(url: str, context: Optional[ExecutionContext] = None) -> 'ContractInterface':
        """Create contract from michelson source code available via URL

        :param url: link to the Michelson file
        :param context: optional execution context
        :rtype: ContractInterface
        """
        res = requests.get(url)
        if res.status_code != 200:
            raise ValueError(f'cannot fetch `{url} {res.status_code}`', res.text)
        return ContractInterface.from_michelson(res.text, context)

    @staticmethod
    def from_file(path: str, context: Optional[ExecutionContext] = None) -> 'ContractInterface':
        """Create contract from michelson source code stored in a file.

        :param path: Path to the `.tz` file
        :param context: optional execution context
        :rtype: ContractInterface
        """
        with open(expanduser(path)) as f:
            return ContractInterface.from_michelson(f.read(), context)

    @staticmethod
    def from_michelson(source: str, context: Optional[ExecutionContext] = None) -> 'ContractInterface':
        """Create contract from michelson source code.

        :param source: Michelson source code
        :param context: optional execution context
        :rtype: ContractInterface
        """
        return ContractInterface.from_micheline(michelson_to_micheline(source), context)

    @staticmethod
    def from_micheline(
        expression: List[Dict[str, Any]],
        context: Optional[ExecutionContext] = None,
    ) -> 'ContractInterface':
        """Create contract from micheline expression.

        :param expression: [{'prim': 'parameter'}, {'prim': 'storage'}, {'prim': 'code'}]
        :param context: optional execution context
        :rtype: ContractInterface
        """
        if context is not None:
            code_expr = context.resolve_global_constants(expression)
        else:
            code_expr = expression
        program = MichelsonProgram.match(code_expr)
        cls = type(ContractInterface.__name__, (ContractInterface,), {'program': program})
        context = ExecutionContext(
            shell=context.shell if context else None,
            key=context.key if context else None,
            script={'code': code_expr},
            global_constants=context.global_constants if context else None,
        )
        return cls(context)

    @staticmethod
    def from_context(context: ExecutionContext) -> 'ContractInterface':
        """Create contract from the previously loaded context data.

        :param context: execution context
        :return: ContractInterface
        """
        program = MichelsonProgram.load(context, with_code=True)
        cls = type(ContractInterface.__name__, (ContractInterface,), {'program': program})
        return cls(context)

    @classmethod
    @deprecated(
        deprecated_in='3.0.0',
        removed_in='4.0.0',
        details='use one of `from_file`, `from_michelson`, `from_micheline`, `from_url`',
    )
    def create_from(cls, source):
        """Create contract interface from its code.

        :param source: Michelson code, filename, or Micheline JSON
        :rtype: ContractInterface
        """
        if isinstance(source, str):
            if exists(expanduser(source)):
                return ContractInterface.from_file(source)
            return ContractInterface.from_michelson(source)
        return ContractInterface.from_micheline(source)

    def to_micheline(self) -> List[Dict[str, Any]]:
        """Get contract script in Micheline JSON

        :return:  [{'prim': 'parameter'}, {'prim': 'storage'}, {'prim': 'code'}]
        """
        return self.program.as_micheline_expr()

    def to_michelson(self) -> str:
        """Get contract listing in formatted Michelson

        :return: string
        """
        return micheline_to_michelson(self.to_micheline())

    def to_file(self, path: str) -> None:
        """Write contract source to a .tz file

        :param path: path to the file
        """
        with open(path, 'w+') as f:
            f.write(self.to_michelson())

    @deprecated(deprecated_in='3.0.0', removed_in='4.0.0', details='use `.storage[path][to][big_map][key]()` instead')
    def big_map_get(self, path):
        """Get BigMap entry as Python object by plain key and block height.

        :param path: JSON path to the key (or just key to access default BigMap location).
            Use `/` to separate nodes and `::` to separate tuple args.
            In any other case you'd need to escape those symbols.
        :returns: object
        """
        node = self.storage
        for item in path.split('/'):
            if len(item) == 0:
                continue
            if isinstance(item, str):
                res = item.split('::')
                item = tuple(res) if len(res) > 1 else item
            node = node[item]
        return node() if node else None

    def using(
        self,
        shell: Optional[Union[ShellQuery, str]] = None,
        key: Optional[Union[Key, str]] = None,
        block_id: Optional[Union[str, int]] = None,
        mode: Optional[str] = None,
        ipfs_gateway: Optional[str] = None,
    ) -> 'ContractInterface':
        """Change the block at which the current contract is inspected.

        Also, if address is undefined you can specify RPC endpoint, and private key.

        :param shell: one of 'mainnet', '***net', or RPC node uri, or instance of :class:`pytezos.rpc.shell.ShellQuery`
        :param key: base58 encoded key, path to the faucet file, alias from tezos-client, or instance of `Key`
        :param block_id: block height / hash / offset to use, default is `head`
        :param mode: whether to use `readable` or `optimized` encoding for parameters/storage/other
        :param ipfs_gateway: override IPFS gateway URI
        :rtype: ContractInterface
        """
        has_address = self.context.address is not None
        return type(self)(
            self._spawn_context(
                shell=None if has_address else shell,
                key=None if has_address else key,
                address=self.context.address,
                block_id=block_id,
                mode=mode,
                ipfs_gateway=ipfs_gateway,
            )
        )

    @property
    def storage(self) -> ContractData:
        if self._storage:
            return self._storage
        elif self.address:
            expr = self.shell.blocks[self.context.block_id].context.contracts[self.address].storage()
            storage = self.program.storage.from_micheline_value(expr)
            storage.attach_context(self.context)
        else:
            storage = self.program.storage.dummy(self.context)
        return ContractData(self.context, storage.item, title="storage")

    @storage.setter
    def storage(self, storage: ContractData) -> None:
        if self.address:
            raise Exception('Can\'t set storage of deployed contract')
        self._storage = storage

    def storage_from_file(self, path: str) -> None:
        """Load contract storage from file

        :param path: path to .tz file
        """
        with open(path) as file:
            expr = michelson_to_micheline(file.read())
        self.storage_from_micheline(expr)

    def storage_from_micheline(self, expression) -> None:
        """Load contract storage from Micheline expression

        :param expression: Micheline expression
        """
        storage = self.program.storage.from_micheline_value(expression)
        storage.attach_context(self.context)
        self.storage = ContractData(self.context, storage.item, title="storage")

    def storage_from_michelson(self, source: str) -> None:
        """Load contract storage from Michelson code

        :param source: Michelson code
        """
        expr = michelson_to_micheline(source)
        self.storage_from_micheline(expr)

    @cached_property
    def metadata(self) -> Optional[ContractMetadata]:
        """Get TZIP-016 contract metadata, if exists

        :rtype: ContractMetadata
        """
        metadata_url = self.metadata_url
        if metadata_url is None:
            return None

        logger.info('Trying to fetch contract metadata from `%s`', metadata_url)
        parsed_url = urlparse(metadata_url)

        if parsed_url.scheme in ('http', 'https'):
            # NOTE: KT1B34qXVRfQrScEaqjjt6cJ5G8LtVFZ7fSc
            metadata = ContractMetadata.from_url(metadata_url, self.context)

        elif parsed_url.scheme == 'ipfs':
            # NOTE: KT1AFA2mwNUMNd4SsujE1YYp29vd8BZejyKW
            metadata = ContractMetadata.from_ipfs(parsed_url.netloc, self.context)

        elif parsed_url.scheme == 'tezos-storage':
            parts = parsed_url.path.split('/')
            if len(parts) == 1:
                # NOTE: KT1JBThDEqyqrEHimhxoUBCSnsKAqFcuHMkP
                storage = self.storage
            elif len(parts) == 2:
                # NOTE: KT1REEb5VxWRjcHm5GzDMwErMmNFftsE5Gpf
                context = self._spawn_context(address=parsed_url.netloc)
                storage = ContractInterface.from_context(context).storage
            else:
                raise NotImplementedError('Unknown metadata URL scheme')
            metadata_json = json.loads(storage['metadata'][parts[-1]]().decode())
            metadata = ContractMetadata.from_json(metadata_json, self.context)

        elif parsed_url.scheme == 'sha256':
            raise NotImplementedError

        else:
            raise NotImplementedError('Unknown metadata URL scheme')

        return metadata

    @property
    def token_metadata(self) -> ContractTokenMetadataProxy:
        """Get TZIP-021 contract token metadata proxy

        :rtype: ContractTokenMetadataProxy
        """
        return ContractTokenMetadataProxy(self._get_token_metadata)  # type: ignore

    @lru_cache(maxsize=1000)  # noqa: B019
    def _get_token_metadata(self, token_id: int) -> Optional[ContractTokenMetadata]:
        token_metadata = self._get_token_metadata_from_view(token_id)
        if token_metadata is None:
            token_metadata = self._get_token_metadata_from_storage(token_id)
        return token_metadata

    def _get_token_metadata_from_storage(self, token_id: int) -> Optional[ContractTokenMetadata]:
        self._logger.info('Trying to fetch token %s metadata from storage', token_id)
        try:
            token_metadata_url = self.storage['token_metadata'][token_id]['token_info']['']().decode()
        # FIXME: Dirty
        except (KeyError, AssertionError):
            self._logger.info('Storage doesn\'t contain metadata URL for token %s', token_id)
            return None

        self._logger.info('Trying to fetch contract token metadata from `%s`', token_metadata_url)
        parsed_url = urlparse(token_metadata_url)

        if parsed_url.scheme in ('http', 'https'):
            token_metadata = ContractTokenMetadata.from_url(token_metadata_url, self.context)

        elif parsed_url.scheme == 'ipfs':
            token_metadata = ContractTokenMetadata.from_ipfs(parsed_url.netloc, self.context)

        elif parsed_url.scheme == 'tezos-storage':
            parts = parsed_url.path.split('/')
            if len(parts) == 1:
                storage = self.storage
            elif len(parts) == 2:
                context = self._spawn_context(address=parsed_url.netloc)
                storage = ContractInterface.from_context(context).storage
            else:
                raise NotImplementedError('Unknown metadata URL scheme')
            token_metadata_json = json.loads(storage['metadata'][parts[-1]]().decode())
            token_metadata = ContractTokenMetadata.from_json(token_metadata_json, self.context)

        elif parsed_url.scheme == 'sha256':
            raise NotImplementedError

        else:
            raise NotImplementedError('Unknown metadata URL scheme')

        return token_metadata

    def _get_token_metadata_from_view(self, token_id: int) -> Optional[ContractTokenMetadata]:
        self._logger.info('Trying to fetch token %s metadata from off-chain view', token_id)
        try:
            token_metadata_json = cast(ContractMetadata, self.metadata).tokenMetadata(token_id).storage_view()[1]
            return ContractTokenMetadata.from_json(token_metadata_json)
        except KeyError:
            self._logger.info('There\'s no off-chain view named `token_metadata`')
            return None
        except MichelsonRuntimeError:
            self._logger.info('Off-chain view has no token metadata for token_id %s', token_id)
            return None

    @cached_property
    def metadata_url(self) -> Optional[str]:
        try:
            metadata = self.storage.data.find(lambda x: x.field_name == 'metadata' and isinstance(x, BigMapType))
            if metadata is not None:
                metadata_url = metadata['']
                if isinstance(metadata_url, BytesType):
                    return metadata_url.value.decode()
                else:
                    self._logger.info('Empty string key is not found in metadata big map')
            else:
                self._logger.info('Metadata big map not found')
        # FIXME: Dirty
        except (KeyError, AssertionError):
            self._logger.info('Failed to get metadata URI')
        return None

    @property
    def parameter(self) -> ContractEntrypoint:
        root_name = self.program.parameter.root_name
        assert root_name in self.entrypoints, 'root entrypoint is undefined'
        return getattr(self, root_name)

    @property  # type: ignore
    @deprecated(deprecated_in='3.0.0', removed_in='4.0.0', details='access `ContractInterface` directly')
    def contract(self) -> 'ContractInterface':
        return self

    @property  # type: ignore
    @deprecated(deprecated_in='3.0.0', removed_in='4.0.0', details='use `to_michelson()` instead')
    def text(self) -> str:
        return self.to_michelson()

    @property  # type: ignore
    @deprecated(deprecated_in='3.0.0', removed_in='4.0.0', details='use `to_micheline()` instead')
    def code(self):
        return self.to_micheline()

    @property  # type: ignore
    @deprecated(deprecated_in='3.0.0', removed_in='4.0.0', details='use `default()` instead')
    def call(self) -> ContractEntrypoint:
        return self.parameter

    def operation_result(self, operation_group: Dict[str, Any]) -> List[ContractCallResult]:
        """Get operation parameters, and resulting storage as Python objects.
        Can locate operation inside operation groups with multiple contents and/or internal operations.

        :param operation_group: {'branch', 'protocol', 'contents', 'signature'}
        :rtype: ContractCallResult
        """
        return ContractCallResult.from_run_operation(operation_group, context=self.context)

    def script(self, initial_storage=None, mode: Optional[str] = None) -> Dict[str, Any]:
        """Generate script for contract origination.

        :param initial_storage: Python object, leave None to generate default (attach shell/key for smart fill)
        :param mode: whether to use `readable` or `optimized` (or `legacy_optimized`) encoding for initial storage
        :return: {"code": $Micheline, "storage": $Micheline}
        """
        if initial_storage:
            storage = self.program.storage.from_python_object(initial_storage)
        else:
            storage = self.program.storage.dummy(self.context)
        return {
            'code': self.program.as_micheline_expr(),
            'storage': storage.to_micheline_value(mode=mode or self.context.mode, lazy_diff=True),
        }

    def originate(
        self,
        initial_storage=None,
        mode: Optional[str] = None,
        balance: Union[int, Decimal] = 0,
        delegate: Optional[str] = None,
    ) -> OperationGroup:
        """Create an origination operation

        :param initial_storage: Python object, leave None to generate default
        :param mode: whether to use `readable` or `optimized` (or `legacy_optimized`) encoding for initial storage
        :param balance: initial balance
        :param delegate: initial delegator
        :rtype: OperationGroup
        """
        return OperationGroup(context=self._spawn_context()).origination(
            script=self.script(initial_storage, mode=mode),
            balance=balance,
            delegate=delegate,
        )


@deprecated(deprecated_in='3.0.0', removed_in='4.0.0', details='use `ContractInterface` instead')
class Contract(ContractInterface):
    pass



import json
from datetime import datetime
from datetime import time
from os.path import dirname
from os.path import join
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import requests
from attr import dataclass
from jsonschema import validate as jsonschema_validate  # type: ignore

from pytezos.context.impl import ExecutionContext
from pytezos.context.mixin import ContextMixin
from pytezos.contract import converter

with open(join(dirname(__file__), 'token_metadata-schema.json')) as file:
    token_metadata_schema = json.load(file)

# NOTE: "time" is a format not a type
token_metadata_schema['definitions']['format']['properties']['duration']['type'] = "string"
token_metadata_schema['definitions']['asset']['properties']['name'] = {'type': 'string'}
token_metadata_schema['definitions']['asset']['properties']['symbol'] = {'type': 'string'}
token_metadata_schema['definitions']['asset']['properties']['decimals'] = {'type': 'integer'}


@dataclass(kw_only=True, frozen=True)
class Attribute:
    name: str
    value: str
    type: Optional[str] = None


@dataclass(kw_only=True, frozen=True)
class Dimensions:
    value: str
    unit: str


@dataclass(kw_only=True, frozen=True)
class DataRate:
    value: str
    unit: str


@dataclass(kw_only=True, frozen=True)
class Format:
    uri: Optional[str] = None
    hash: Optional[str] = None
    mimeType: Optional[str] = None
    fileSize: Optional[int] = None
    fileName: Optional[str] = None
    duration: Optional[time] = None
    dimensions: Optional[Dimensions] = None
    dataRate: Optional[DataRate] = None


@dataclass(kw_only=True)
class ContractTokenMetadata(ContextMixin):
    """TZIP-21 token metadata"""

    name: Optional[str] = None
    symbol: Optional[str] = None
    decimals: int

    # Fungible Token Recommendations
    symbolPreference: Optional[str] = None
    thumbnailUri: Optional[str] = None

    # Semi-fungible and NFT Token Recommendations
    artifactUri: Optional[str] = None
    displayUri: Optional[str] = None
    description: Optional[str] = None
    minter: Optional[str] = None
    creators: Optional[List[str]] = None
    isBooleanAmount: Optional[bool] = None

    # Multimedia NFT Token Recommendations
    formats: Optional[List[Format]] = None
    tags: Optional[List[str]] = None

    contributors: Optional[List[str]] = None
    publishers: Optional[List[str]] = None
    date: Optional[datetime] = None
    blockLevel: Optional[int] = None
    type: Optional[str] = None
    genres: Optional[List[str]] = None
    language: Optional[List[str]] = None
    identifier: Optional[str] = None
    rights: Optional[str] = None
    rightsUri: Optional[str] = None
    externalUri: Optional[str] = None
    isTransferable: Optional[bool] = None
    shouldPreferSymbol: Optional[str] = None
    attributes: Optional[List[Attribute]] = None
    assets: Optional[List['ContractTokenMetadata']] = None

    def __attrs_post_init__(self):
        self.raw = {}
        self.storage_view_impl = {}

    @staticmethod
    def validate_token_metadata_json(metadata_json: Dict[str, Any]) -> None:
        """Validate token metadata JSON with JSONSchema"""
        jsonschema_validate(instance=metadata_json, schema=token_metadata_schema)

    @classmethod
    def from_json(
        cls,
        token_metadata_json: Dict[str, Any],
        context: Optional[ExecutionContext] = None,
    ) -> 'ContractTokenMetadata':
        """Convert token metadata from JSON object"""

        for key, value in token_metadata_json.items():
            if isinstance(value, bytes):
                token_metadata_json[key] = value.decode()
        if 'decimals' in token_metadata_json:
            token_metadata_json['decimals'] = int(token_metadata_json['decimals'])

        cls.validate_token_metadata_json(token_metadata_json)
        res = converter.structure(token_metadata_json, ContractTokenMetadata)
        res.context = context if context else ExecutionContext()
        res.raw = token_metadata_json
        return res

    @classmethod
    def from_ipfs(cls, multihash: str, context: Optional[ExecutionContext] = None) -> 'ContractTokenMetadata':
        """Fetch token metadata from IPFS network by multihash"""
        context = context or ExecutionContext()
        token_metadata_json = requests.get(f'{context.ipfs_gateway}/{multihash}').json()
        return cls.from_json(token_metadata_json, context)

    @classmethod
    def from_url(cls, url: str, context: Optional[ExecutionContext] = None) -> 'ContractTokenMetadata':
        """Fetch token metadata from HTTP(S) URL"""
        token_metadata_json = requests.get(url).json()
        return cls.from_json(token_metadata_json, context)

    @classmethod
    def from_file(cls, path: str, context: Optional[ExecutionContext] = None) -> 'ContractTokenMetadata':
        """Read token metadata from JSON file by path"""
        with open(path) as f:
            token_metadata_json = json.load(f)
            return cls.from_json(token_metadata_json, context)



from typing import List

from pyblake2 import blake2b  # type: ignore

from pytezos.crypto.encoding import base58_decode
from pytezos.crypto.encoding import base58_encode


def _hash_tuple(left: bytes = b'', right: bytes = b'') -> bytes:
    return blake2b(left + right, digest_size=32).digest()


def _reduce_operation_hashes(hashes: List[bytes]) -> bytes:
    a: List[bytes] = []

    def step(n: int) -> bytes:
        nonlocal a
        m = (n + 1) // 2
        for i in range(m):
            a[i] = _hash_tuple(a[2 * i], a[2 * i + 1])
        a[m] = _hash_tuple(a[n], a[n])
        if m == 1:
            return a[0]
        elif m % 2 == 0:
            return step(m)
        else:
            a[m + 1] = a[m]
            return step(m + 1)

    if len(hashes) == 0:
        return _hash_tuple()
    elif len(hashes) == 1:
        return _hash_tuple(hashes[0])
    else:
        res = list(map(lambda x: _hash_tuple(x), hashes))
        a = res + [res[-1]]
        return step(len(hashes))


def operation_list_hash(operation_hashes: List[str]) -> str:
    raw_items = list(map(lambda x: base58_decode(x.encode()), operation_hashes))
    res = _reduce_operation_hashes(raw_items)
    return base58_encode(res, b'Lo').decode()


def operation_list_list_hash(operations_hashes: List[List[str]]) -> str:
    lo_hashes = list(map(operation_list_hash, operations_hashes))
    raw_items = list(map(lambda x: base58_decode(x.encode()), lo_hashes))
    res = _reduce_operation_hashes(raw_items)
    return base58_encode(res, b'LLo').decode()


def block_payload_hash(predecessor: str, payload_round: int, operation_hashes: List[str]) -> str:
    """Calculate payload hash
    For each level, Tenderbake proceeds in rounds. Each round represents an attempt by the validators to agree on the
    content of the block for the current level, that is, on the sequence of non-consensus operations the block contains.
    We call this sequence the block’s payload.
    :param predecessor: block hash (base58 encoded) of the previous block
    :param payload_round: round number (int32)
    :param operation_hashes: flat list of non-consensus (validation pass > 0) operation hashes
    """
    # https://gitlab.com/tezos/tezos/-/blob/master/src/proto_alpha/lib_delegate/block_forge.ml#L166
    # https://gitlab.com/tezos/tezos/-/blob/master/src/proto_012_Psithaca/lib_protocol/block_payload_repr.ml#L41
    payload = [
        base58_decode(predecessor.encode()),
        payload_round.to_bytes(4, 'big'),
        _reduce_operation_hashes([base58_decode(x.encode()) for x in operation_hashes]),
    ]
    res = blake2b(b''.join(payload), digest_size=32).digest()
    return base58_encode(res, b'vh').decode()



from typing import Any
from typing import Optional
from typing import Type

from pytezos.michelson.micheline import MichelineSequence
from pytezos.michelson.types.base import MichelsonType


class OperationType(MichelsonType, prim='operation'):
    def __init__(self, content: dict, ty: Optional[Type[MichelsonType]] = None):
        super(OperationType, self).__init__()
        self.content = content
        self.ty = ty

    def __repr__(self):
        return self.content['kind']

    def __eq__(self, other):
        if not isinstance(other, OperationType):
            return False
        return self.content == other.content

    @classmethod
    def origination(
        cls,
        source: str,
        script: Type[MichelineSequence],
        storage: MichelsonType,
        balance: int = 0,
        delegate: Optional[str] = None,
    ) -> 'OperationType':
        content = {
            'kind': 'origination',
            'source': source,
            'script': {
                'code': script.as_micheline_expr(),
                'storage': storage.to_micheline_value(),
            },
            'balance': str(balance),
        }
        if delegate is not None:
            content['delegate'] = delegate
        return cls(content, ty=type(storage))

    @classmethod
    def delegation(cls, source: str, delegate: Optional[str] = None) -> 'OperationType':
        content = {
            'kind': 'delegation',
            'source': source,
            'delegate': delegate,
        }
        return cls(content)

    @classmethod
    def transaction(
        cls,
        source: str,
        destination: str,
        amount: int,
        entrypoint: str,
        value: Any,
        param_type: Type[MichelsonType],
    ) -> 'OperationType':
        content = {
            'kind': 'transaction',
            'source': source,
            'destination': destination,
            'amount': str(amount),
            'parameters': {
                'entrypoint': entrypoint,
                'value': value,
            },
        }
        return cls(content, ty=param_type)

    @classmethod
    def event(cls, source: str, event_type: Type[MichelsonType], payload: Any, tag: str) -> 'OperationType':
        content = {
            'kind': 'event',
            'source': source,
            'event_type': event_type.as_micheline_expr(),
            'payload': payload,
            'tag': tag,
        }
        return cls(content, ty=event_type)

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        kind = self.content['kind']
        assert self.ty, f'data type is not defined for {kind}'
        if kind == 'transaction':
            data = self.ty.from_micheline_value(self.content['parameters']['value'])
        elif kind == 'origination':
            data = self.ty.from_micheline_value(self.content['script']['storage'])
        elif kind == 'event':
            data = self.ty.from_micheline_value(self.content['payload'])
        else:
            raise AssertionError(f'not applicable for {kind}')
        return data.to_python_object(try_unpack=try_unpack)



from typing import List
from typing import Tuple
from typing import Type
from typing import Union
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.instructions.adt import PairInstruction
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.base import Wildcard
from pytezos.michelson.instructions.base import format_stdout
from pytezos.michelson.instructions.stack import PushInstruction
from pytezos.michelson.micheline import MichelineSequence
from pytezos.michelson.micheline import MichelsonRuntimeError
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types import BoolType
from pytezos.michelson.types import LambdaType
from pytezos.michelson.types import ListType
from pytezos.michelson.types import MapType
from pytezos.michelson.types import MichelsonType
from pytezos.michelson.types import OptionType
from pytezos.michelson.types import OrType
from pytezos.michelson.types import PairType
from pytezos.michelson.types import SetType


def execute_dip(
    prim: str,
    stack: MichelsonStack,
    stdout: List[str],
    count: int,
    body: Type[MichelsonInstruction],
    context: AbstractContext,
) -> MichelsonInstruction:
    stdout.append(format_stdout(prim, [*Wildcard.n(count)], []))
    stack.protect(count=count)
    item = body.execute(stack, stdout, context=context)
    stack.restore(count=count)
    stdout.append(format_stdout(prim, [], [*Wildcard.n(count)], count))
    return item


class DipnInstruction(MichelsonInstruction, prim='DIP', args_len=2):
    def __init__(self, item: MichelsonInstruction):
        super(DipnInstruction, self).__init__()
        self.item = item

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        depth = cls.args[0].get_int()  # type: ignore
        item = execute_dip(cls.prim, stack, stdout, count=depth, body=cls.args[1], context=context)  # type: ignore
        return cls(item)


class DipInstruction(MichelsonInstruction, prim='DIP', args_len=1):
    def __init__(self, item: MichelsonInstruction):
        super(DipInstruction, self).__init__()
        self.item = item

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        item = execute_dip(cls.prim, stack, stdout, count=1, body=cls.args[0], context=context)  # type: ignore
        return cls(item)


class LambdaInstruction(MichelsonInstruction, prim='LAMBDA', args_len=3):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        lambda_type = LambdaType.create_type(args=cls.args[:2])
        res = lambda_type(cls.args[2])  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [], [res]))  # type: ignore
        return cls(stack_items_added=1)


class ExecInstruction(MichelsonInstruction, prim='EXEC'):
    def __init__(self, item: MichelsonInstruction):
        super(ExecInstruction, self).__init__(stack_items_added=1)
        self.item = item

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        param, lambda_ = cast(Tuple[MichelsonType, LambdaType], stack.pop2())
        assert isinstance(lambda_, LambdaType), f'expected lambda, got {lambda_.prim}'
        param.assert_type_equal(lambda_.args[0])
        stdout.append(format_stdout(cls.prim, [param, lambda_], []))  # type: ignore
        lambda_stack = MichelsonStack.from_items([param])
        lambda_body = cast(MichelsonInstruction, lambda_.value)
        item = lambda_body.execute(lambda_stack, stdout, context=context)
        res = lambda_stack.pop1()
        res.assert_type_equal(lambda_.args[1])
        assert len(lambda_stack) == 0, f'lambda stack is not empty {lambda_stack}'
        stack.push(res)
        return cls(item)


class ApplyInstruction(MichelsonInstruction, prim='APPLY'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        left, lambda_ = cast(Tuple[MichelsonType, LambdaType], stack.pop2())
        lambda_.assert_type_in(LambdaType)
        lambda_.args[0].assert_type_in(PairType)
        left_type, right_type = lambda_.args[0].args
        left.assert_type_equal(left_type)

        new_value = MichelineSequence.create_type(
            args=[
                PushInstruction.create_type(args=[left_type, left.to_literal()]),
                PairInstruction,
                lambda_.value,
            ]
        )
        res = LambdaType.create_type(args=[right_type, lambda_.args[1]])(new_value)  # type: ignore
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [left, lambda_], [res]))  # type: ignore
        return cls(stack_items_added=1)


class FailwithInstruction(MichelsonInstruction, prim='FAILWITH'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a = stack.pop1()
        assert a.is_packable(), f'expected packable type, got {a.prim}'
        raise MichelsonRuntimeError(repr(a))


class IfInstruction(MichelsonInstruction, prim='IF', args_len=2):
    def __init__(self, item: MichelsonInstruction):
        super(IfInstruction, self).__init__()
        self.item = item

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        cond = cast(BoolType, stack.pop1())
        cond.assert_type_equal(BoolType)
        stdout.append(format_stdout(cls.prim, [cond], []))  # type: ignore
        branch = cls.args[0] if bool(cond) else cls.args[1]
        item = branch.execute(stack, stdout, context=context)
        return cls(item)


class IfConsInstruction(MichelsonInstruction, prim='IF_CONS', args_len=2):
    def __init__(self, stack_items_added: int, item: MichelsonInstruction):
        super(IfConsInstruction, self).__init__(stack_items_added)
        self.item = item

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        lst = cast(ListType, stack.pop1())
        lst.assert_type_in(ListType)
        if len(lst) > 0:
            head, tail = lst.split_head()
            stack.push(tail)
            stack.push(head)
            stdout.append(format_stdout(cls.prim, [lst], [head, tail]))  # type: ignore
            branch = cls.args[0]
            stack_items_added = 2
        else:
            stdout.append(format_stdout(cls.prim, [lst], []))  # type: ignore
            branch = cls.args[1]
            stack_items_added = 0
        item = branch.execute(stack, stdout, context=context)
        return cls(stack_items_added, item)


class IfLeftInstruction(MichelsonInstruction, prim='IF_LEFT', args_len=2):
    def __init__(self, item: MichelsonInstruction):
        super(IfLeftInstruction, self).__init__(stack_items_added=1)
        self.item = item

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        or_ = cast(OrType, stack.pop1())
        or_.assert_type_in(OrType)
        branch = cls.args[0] if or_.is_left() else cls.args[1]
        res = or_.resolve()
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [or_], [res]))  # type: ignore
        item = branch.execute(stack, stdout, context=context)
        return cls(item)


class IfNoneInstruction(MichelsonInstruction, prim='IF_NONE', args_len=2):
    def __init__(self, stack_items_added: int, item: MichelsonInstruction):
        super(IfNoneInstruction, self).__init__(stack_items_added)
        self.item = item

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        opt = cast(OptionType, stack.pop1())
        opt.assert_type_in(OptionType)
        if opt.is_none():
            branch = cls.args[0]
            stdout.append(format_stdout(cls.prim, [opt], []))  # type: ignore
            stack_items_added = 0
        else:
            some = opt.get_some()
            stack.push(some)
            stdout.append(format_stdout(cls.prim, [opt], [some]))  # type: ignore
            branch = cls.args[1]
            stack_items_added = 1
        item = branch.execute(stack, stdout, context=context)
        return cls(stack_items_added, item)


class LoopInstruction(MichelsonInstruction, prim='LOOP', args_len=1):
    def __init__(self, items: List[MichelsonInstruction]):
        super(LoopInstruction, self).__init__()
        self.items = items

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        items = []
        while True:
            cond = cast(BoolType, stack.pop1())
            cond.assert_type_equal(BoolType)
            stdout.append(format_stdout(cls.prim, [cond], []))  # type: ignore
            if bool(cond):
                item = cls.args[0].execute(stack, stdout, context=context)
                items.append(item)
            else:
                break
        return cls(items)


class LoopLeftInstruction(MichelsonInstruction, prim='LOOP_LEFT', args_len=1):
    def __init__(self, stack_items_added: int, items: List[MichelsonInstruction]):
        super(LoopLeftInstruction, self).__init__(stack_items_added)
        self.items = items

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        stack_items_added = 0
        items = []
        while True:
            or_ = cast(OrType, stack.pop1())
            or_.assert_type_in(OrType)
            var = or_.resolve()
            stack.push(var)
            stack_items_added += 1
            stdout.append(format_stdout(cls.prim, [or_], [var]))  # type: ignore
            if or_.is_left():
                item = cls.args[0].execute(stack, stdout, context=context)
                items.append(item)
            else:
                break
        return cls(stack_items_added, items)


class MapInstruction(MichelsonInstruction, prim='MAP', args_len=1):
    def __init__(self, stack_items_added: int, items: List[MichelsonInstruction]):
        super(MapInstruction, self).__init__(stack_items_added)
        self.items = items

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        stack_items_added = 0  # noqa: SIM113
        src = cast(Union[ListType, MapType], stack.pop1())
        executions = []
        items = []
        popped = [src]
        for elt in src:
            if isinstance(src, MapType):
                elt = PairType.from_comb(list(elt))  # type: ignore
            stack.push(elt)  # type: ignore
            stack_items_added += 1
            stdout.append(format_stdout(cls.prim, popped, [elt]))  # type: ignore
            execution = cls.args[0].execute(stack, stdout, context=context)
            executions.append(execution)
            new_elt = stack.pop1()
            if isinstance(src, MapType):
                items.append((elt[0], new_elt))
            else:
                items.append(new_elt)  # type: ignore
            popped = [new_elt]  # type: ignore

        if items:
            res = type(src).from_items(items)  # type: ignore
        else:
            res = src  # TODO: need to deduce argument types
        stack.push(res)
        stack_items_added += 1
        stdout.append(format_stdout(cls.prim, popped, [res]))  # type: ignore
        return cls(stack_items_added, executions)


class IterInstruction(MichelsonInstruction, prim='ITER', args_len=1):
    def __init__(self, stack_items_added: int, items: List[MichelsonInstruction]):
        super(IterInstruction, self).__init__(stack_items_added)
        self.items = items

    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        stack_items_added = 0  # noqa: SIM113
        src = cast(Union[ListType, MapType, SetType], stack.pop1())
        executions = []
        popped = [src]
        for elt in src:
            if isinstance(src, MapType):
                elt = PairType.from_comb(list(elt))  # type: ignore
            stack_items_added += 1
            stack.push(elt)  # type: ignore
            stdout.append(format_stdout(cls.prim, popped, [elt]))  # type: ignore
            execution = cls.args[0].execute(stack, stdout, context=context)
            executions.append(execution)
            popped = []
        return cls(stack_items_added, executions)



from typing import Optional
from typing import Union

from deprecation import deprecated  # type: ignore

from pytezos.context.impl import ExecutionContext
from pytezos.context.mixin import ContextMixin
from pytezos.jupyter import get_class_docstring
from pytezos.michelson.format import micheline_to_michelson
from pytezos.michelson.parse import michelson_to_micheline
from pytezos.michelson.types.base import MichelsonType
from pytezos.michelson.types.base import generate_pydoc


class ContractData(ContextMixin):
    def __init__(self, context: ExecutionContext, data: MichelsonType, path='', title=None) -> None:
        super().__init__(context=context)
        self.data = data
        self.path = path
        self.__doc__ = generate_pydoc(type(self.data), title=title)

    def __repr__(self) -> str:
        res = [
            super().__repr__(),
            f'.path\t{self.path}',
            f'\nBuiltin\n()\t# get as Python object',
            f'[key]  # access child elements by name or index',
            f'\nTypedef\n{self.__doc__}',
            '\nHelpers',
            get_class_docstring(self.__class__),
        ]
        return '\n'.join(res)

    def __getitem__(self, item: Union[str, int]) -> 'ContractData':
        """Access child elements by name or index (depending on the type)

        :param item: field name (str) or index (int)
        :rtype: ContractData
        """
        res = self.data[item]
        if res is None:
            raise KeyError(item)
        return ContractData(self.context, res, path=f'{self.path}/{item}')

    def __call__(self, try_unpack=False):
        """Get Michelson value as a Python object

        :param try_unpack: try to unpack utf8-encoded strings or PACKed Michelson expressions
        """
        return self.data.to_python_object(try_unpack=try_unpack)

    def to_micheline(self, optimized=False):
        """Get as Micheline JSON expression

        :param optimized: use optimized data form for some domain types (timestamp, address, etc.)
        """
        return self.data.to_micheline_value(mode='optimized' if optimized else 'readable', lazy_diff=None)

    def to_michelson(self, optimized=False):
        """Get as Michelson value

        :param optimized: use optimized data form for some domain types (timestamp, address, etc.)
        """
        return micheline_to_michelson(self.to_micheline(optimized=optimized))

    def decode(self, value):
        """Convert from Michelson or Micheline to Python type system

        :param value: Micheline JSON expression or Michelson value
        :return: Python object
        """
        if isinstance(value, str):
            value = michelson_to_micheline(value)
        return type(self.data).from_micheline_value(value).to_python_object(lazy_diff=None)

    def encode(self, py_obj, mode: Optional[str] = None):
        """Convert from Python to Micheline type system

        :param py_obj: Python object
        :param mode: whether to use `readable` or `optimized` (or `legacy_optimized`) encoding
        :return: Micheline JSON expression
        """
        return (
            type(self.data)
            .from_python_object(py_obj)
            .to_micheline_value(mode=mode or self.context.mode, lazy_diff=None)
        )

    def dummy(self):
        """Try to generate a dummy (empty) value

        :return: Python object
        """
        return type(self.data).dummy(self.context).to_python_object(lazy_diff=True)

    @deprecated(deprecated_in='3.0.0', removed_in='4.0.0')
    def default(self):
        return self.dummy()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['michelson_kernel',
 'pytezos',
 'pytezos.block',
 'pytezos.cli',
 'pytezos.context',
 'pytezos.contract',
 'pytezos.crypto',
 'pytezos.michelson',
 'pytezos.michelson.instructions',
 'pytezos.michelson.sections',
 'pytezos.michelson.types',
 'pytezos.operation',
 'pytezos.protocol',
 'pytezos.rpc',
 'pytezos.sandbox']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21.4.0,<22.0.0',
 'base58>=2.1.1,<3.0.0',
 'bson>=0.5.10,<0.6.0',
 'cattrs>=22.1.0,<23.0.0',
 'click>=8.0.3,<=8.0.4',
 'deprecation>=2.1.0,<3.0.0',
 'docker>=6.0.0,<7.0.0',
 'exceptiongroup>=1.0.0-rc.8,<2.0.0',
 'fastecdsa>=2.2.3,<3.0.0',
 'ipykernel>=6.6.0,<7.0.0',
 'jsonschema>=4.3.2,<5.0.0',
 'jupyter-client>=7.1.0,<8.0.0',
 'mnemonic>=0.20,<0.21',
 'netstruct>=1.1.2,<2.0.0',
 'notebook>=6.4.6,<7.0.0',
 'pendulum>=2.1.2,<3.0.0',
 'ply>=3.11,<4.0',
 'py-ecc>=6.0.0,<7.0.0',
 'pyblake2>=1.1.2,<2.0.0',
 'pysha3>=1.0.2,<2.0.0',
 'pysodium>=0.7.10,<0.8.0',
 'python-dateutil>=2.8.2,<3.0.0',
 'requests>=2.26.0,<3.0.0',
 'secp256k1>=0.14.0,<0.15.0',
 'simplejson>=3.17.6,<4.0.0',
 'strict-rfc3339>=0.7,<0.8',
 'tabulate>=0.8.9,<0.9.0',
 'testcontainers>=3.4.2,<4.0.0',
 'tqdm>=4.62.3,<5.0.0']

entry_points = \
{'console_scripts': ['michelson-kernel = michelson_kernel.cli:cli',
                     'pytezos = pytezos.cli.cli:cli']}

setup_kwargs = {
    'name': 'pytezos',
    'version': '3.7.1',
    'description': 'Python toolkit for Tezos',
    'long_description': '# PyTezos\n\n[![PyPI version](https://badge.fury.io/py/pytezos.svg?)](https://badge.fury.io/py/pytezos)\n[![Tests](https://github.com/baking-bad/pytezos/workflows/Tests/badge.svg?)](https://github.com/baking-bad/pytezos/actions?query=workflow%3ATests)\n[![Docker images](https://github.com/baking-bad/pytezos/workflows/Dockerhub/badge.svg?)](https://hub.docker.com/r/bakingbad/pytezos)\n[![Made With](https://img.shields.io/badge/made%20with-python-blue.svg?)](ttps://www.python.org)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/baking-bad/pytezos/master?filepath=michelson_quickstart.ipynb)\n\n\n* RPC query engine\n* Cryptography\n* Building and parsing operations\n* Smart contract interaction\n* Local forging/packing & vice versa\n* Working with Michelson AST\n\n#### PyTezos CLI\n* Generating contract parameter/storage schema\n* Activating and revealing accounts\n* Deploying contracts (+ GitHub integration)\n\n#### Michelson REPL\n* Builtin interpreter (reimplemented)\n* Set of extra helpers (stack visualization, blockchain context mocking)\n\n#### Michelson Jupyter kernel\n* Custom interpreter with runtime type checker\n* Syntax highlighting, autocomplete with `Tab`\n* In-place docstrings with `Shift+Tab`\n* Macros support\n* Verbose execution logging\n* Debug helpers\n\n#### Michelson integration testing framework\n* Writing integration tests using `unittest` package\n* Simulating contract execution using remote intepreter (via RPC) or builtin one\n\n\n## Installation\n\nYou need to install cryptographic packages before installing the library/building the project:\n\n#### Linux\n\n##### Ubuntu, Debian and other apt-based distributions\n```shell\n$ sudo apt install libsodium-dev libsecp256k1-dev libgmp-dev pkg-config\n```\n\n##### Arch Linux\n```shell\n$ sudo pacman -Syu --needed libsodium libsecp256k1 gmp\n```\n#### MacOS\n\n[Homebrew](https://brew.sh/) needs to be installed.\n```shell\n$ brew tap cuber/homebrew-libsecp256k1\n$ brew install libsodium libsecp256k1 gmp pkg-config\n```\n\n##### M1 (ARM)\n\nIn case `secp256k1` or `gmp` cannot find either include or lib paths, try explicitly set environment vars:\n```\nCFLAGS="-I/opt/homebrew/Cellar/gmp/6.2.1_1/include/ -L/opt/homebrew/Cellar/gmp/6.2.1_1/lib/" LIB_DIR="/opt/homebrew/Cellar/libsecp256k1/0.1/lib" INCLUDE_DIR=/opt/homebrew/Cellar/libsecp256k1/0.1/include pip3 install --user pytezos\n```\n\nFor running tests you might also need to export `LD_LIBRARY_PATH`:\n```\nexport LD_LIBRARY_PATH=/opt/homebrew/lib/\n```\n\n#### Windows\n\nThe recommended way is to use WSL and then follow the instructions for Linux,\nbut if you feel lucky you can try to install natively:\n\n1. Install MinGW from [https://osdn.net/projects/mingw/](https://osdn.net/projects/mingw/)\n2. Make sure `C:\\MinGW\\bin` is added to your `PATH`\n3. Download the latest libsodium-X.Y.Z-msvc.zip from [https://download.libsodium.org/libsodium/releases/](https://download.libsodium.org/libsodium/releases/).\n4. Extract the Win64/Release/v143/dynamic/libsodium.dll from the zip file\n5. Copy libsodium.dll to C:\\Windows\\System32\\libsodium.dll\n\n### From PyPi\n\n```shell\n$ pip install wheel setuptools pkginfo cryptography\n$ pip install pytezos\n```\n\n### [Google Colab](https://colab.research.google.com)\n\n`````python\n>>> !apt install libsodium-dev libsecp256k1-dev libgmp-dev\n>>> !pip install pytezos\n`````\n\n### Docker container\nVerified & minified images for CI/CD https://hub.docker.com/r/bakingbad/pytezos/tags\n```shell\n$ # 1. Use image from registry\n$ docker pull bakingbad/pytezos\n$ # or build it yourself\n$ docker build . -t pytezos\n$ # 2. Use included docker-compose.yml\n$ docker-compose up -d notebook\n```\n\n### Building from sources\n\nRequirements:\n* Python 3.8+\n* [Poetry](https://python-poetry.org/docs/#installation)\n* libsodium, libsecp256k1, gmp\n* make\n\n```shell\n$ # prepare environment\n$ make install\n# # run full CI with tests\n$ make\n```\n\n## Quick start\nRead [quick start guide](https://pytezos.org/quick_start.html)  \nLearn how to [enable Jupyter with Michelson](./src/michelson_kernel/README.md)\n\n## API reference\nCheck out a complete [API reference](https://pytezos.org/contents.html)\n\n### Inline documentation\nIf you are working in Jupyter/Google Colab or any other interactive console, \nyou can display documentation for a particular class/method:\n\n```python\n>>> from pytezos import pytezos\n>>> pytezos\n```\n\n### Publications\n\n* Pytezos 2.0 release with embedded docs and smart contract interaction engine  \nhttps://medium.com/coinmonks/high-level-interface-for-michelson-contracts-and-not-only-7264db76d7ae\n\n* Materials from TQuorum:Berlin workshop - building an app on top of PyTezos and ConseilPy  \nhttps://medium.com/coinmonks/atomic-tips-berlin-workshop-materials-c5c8ee3f46aa\n\n* Materials from the EETH hackathon - setting up a local development infrastructure, deploying and interacting with a contract  \nhttps://medium.com/tezoscommons/preparing-for-the-tezos-hackathon-with-baking-bad-45f2d5fca519\n\n* Introducing integration testing engine  \nhttps://medium.com/tezoscommons/testing-michelson-contracts-with-pytezos-513718499e93\n\n### Contact\n* Telegram chat: [@baking_bad_chat](https://t.me/baking_bad_chat)\n* Slack channel: [#baking-bad](https://tezos-dev.slack.com/archives/CV5NX7F2L)\n\n## Credits\n* The project was initially started by Arthur Breitman, now it\'s maintained by Baking Bad team.\n* Baking Bad is supported by Tezos Foundation\n* Michelson test set from the Tezos repo is used to ensure the interpreter workability\n* Michelson structured documentation by Nomadic Labs is used for inline help\n',
    'author': 'Michael Zaikin',
    'author_email': 'mz@baking-bad.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://pytezos.org',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from typing import Callable
from typing import List
from typing import Tuple
from typing import Union
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.instructions.base import MichelsonInstruction
from pytezos.michelson.instructions.base import dispatch_types
from pytezos.michelson.instructions.base import format_stdout
from pytezos.michelson.stack import MichelsonStack
from pytezos.michelson.types import BoolType
from pytezos.michelson.types import IntType
from pytezos.michelson.types import NatType


def execute_boolean_add(prim: str, stack: MichelsonStack, stdout: List[str], add: Callable):
    a, b = cast(Tuple[Union[BoolType, NatType], ...], stack.pop2())
    res_type, convert = dispatch_types(
        type(a),
        type(b),
        mapping={
            (BoolType, BoolType): (BoolType, bool),
            (NatType, NatType): (NatType, int),
        },
    )
    val = add((convert(a), convert(b)))
    res = res_type.from_value(val)
    stack.push(res)
    stdout.append(format_stdout(prim, [a, b], [res]))


class OrInstruction(MichelsonInstruction, prim='OR'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        execute_boolean_add(cls.prim, stack, stdout, lambda x: x[0] | x[1])  # type: ignore
        return cls(stack_items_added=1)


class XorInstruction(MichelsonInstruction, prim='XOR'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        execute_boolean_add(cls.prim, stack, stdout, lambda x: x[0] ^ x[1])  # type: ignore
        return cls(stack_items_added=1)


class AndInstruction(MichelsonInstruction, prim='AND'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a, b = cast(Tuple[Union[BoolType, NatType, IntType], ...], stack.pop2())
        res_type, convert = dispatch_types(
            type(a),
            type(b),
            mapping={
                (BoolType, BoolType): (BoolType, bool),
                (NatType, NatType): (NatType, int),
                (NatType, IntType): (NatType, int),
                (IntType, NatType): (NatType, int),
            },
        )
        res = res_type.from_value(convert(a) & convert(b))
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [a, b], [res]))  # type: ignore
        return cls(stack_items_added=1)


class NotInstruction(MichelsonInstruction, prim='NOT'):
    @classmethod
    def execute(cls, stack: MichelsonStack, stdout: List[str], context: AbstractContext):
        a = cast(Union[IntType, NatType, BoolType], stack.pop1())
        res_type, convert = dispatch_types(
            type(a),
            mapping={
                (NatType,): (IntType, lambda x: ~int(x)),
                (IntType,): (IntType, lambda x: ~int(x)),
                (BoolType,): (BoolType, lambda x: not bool(x)),
            },
        )
        res = res_type.from_value(convert(a))
        stack.push(res)
        stdout.append(format_stdout(cls.prim, [a], [res]))  # type: ignore
        return cls(stack_items_added=1)



from typing import Callable
from typing import List
from typing import Optional
from typing import Type

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.micheline import parse_micheline_value
from pytezos.michelson.types.base import MichelsonType


class NoneLiteral(Micheline, prim='None'):
    pass


class SomeLiteral(Micheline, prim='Some', args_len=1):
    pass


class OptionType(MichelsonType, prim='option', args_len=1):
    def __init__(self, item: Optional[MichelsonType]):
        super(OptionType, self).__init__()
        self.item = item

    def __lt__(self, other: 'OptionType') -> bool:  # type: ignore
        if other.item is None:
            return False
        elif self.item is None:
            return True
        else:
            return self.item < other.item

    def __eq__(self, other) -> bool:  # type: ignore
        if not isinstance(other, OptionType):
            return False
        return self.item == other.item  # type: ignore

    def __hash__(self):
        return hash(self.item)

    def __repr__(self):
        return f'{repr(self.item)}?' if self.item else 'None'

    @staticmethod
    def none(some_type: Type[MichelsonType]) -> 'OptionType':
        cls = OptionType.create_type(args=[some_type])
        return cls(None)  # type: ignore

    @staticmethod
    def from_some(item: MichelsonType) -> 'OptionType':
        cls = OptionType.create_type(args=[item.get_anon_type()])
        return cls(item)  # type: ignore

    @classmethod
    def dummy(cls, context: AbstractContext) -> 'OptionType':
        return cls(None)

    @classmethod
    def generate_pydoc(cls, definitions: list, inferred_name=None, comparable=False):
        name = cls.field_name or cls.type_name or inferred_name
        arg_doc = cls.args[0].generate_pydoc(definitions, inferred_name=name)
        return f'{arg_doc} || None'

    @classmethod
    def from_micheline_value(cls, val_expr):
        item = parse_micheline_value(
            val_expr,
            {
                ('Some', 1): lambda x: cls.args[0].from_micheline_value(x[0]),
                ('None', 0): lambda x: None,
            },
        )
        return cls(item)

    @classmethod
    def from_python_object(cls, py_obj):
        if py_obj is None:
            item = None
        else:
            item = cls.args[0].from_python_object(py_obj)
        return cls(item)

    def is_none(self) -> bool:
        return self.item is None

    def get_some(self) -> MichelsonType:
        assert not self.is_none()
        return self.item  # type: ignore

    def to_literal(self) -> Type[Micheline]:
        if self.is_none():
            return NoneLiteral
        else:
            return SomeLiteral.create_type(args=[self.item.to_literal()])  # type: ignore

    def to_micheline_value(self, mode='readable', lazy_diff=False):
        if self.is_none():
            return {'prim': 'None'}
        else:
            arg = self.item.to_micheline_value(mode=mode, lazy_diff=lazy_diff)
            return {'prim': 'Some', 'args': [arg]}

    def to_python_object(self, try_unpack=False, lazy_diff=False, comparable=False):
        if self.is_none():
            return None
        else:
            return self.item.to_python_object(
                try_unpack=try_unpack,
                lazy_diff=lazy_diff,
                comparable=comparable,
            )

    def merge_lazy_diff(self, lazy_diff: List[dict]) -> 'MichelsonType':
        item = None if self.is_none() else self.item.merge_lazy_diff(lazy_diff)  # type: ignore
        return type(self)(item)

    def aggregate_lazy_diff(self, lazy_diff: List[dict], mode='readable') -> 'MichelsonType':
        item = None if self.is_none() else self.item.aggregate_lazy_diff(lazy_diff, mode=mode)  # type: ignore
        return type(self)(item)

    def attach_context(self, context: AbstractContext, big_map_copy=False):
        if not self.is_none():
            self.item.attach_context(context, big_map_copy=big_map_copy)  # type: ignore

    def find(self, predicate: Callable[['MichelsonType'], bool]) -> Optional['MichelsonType']:
        if predicate(self):
            return self
        if self.item is not None:
            return self.item.find(predicate)
        return None



import io
import logging
import sys
import tarfile
import time
from glob import glob
from os.path import abspath
from os.path import dirname
from os.path import exists
from os.path import join
from os.path import split
from pprint import pformat
from typing import List
from typing import Optional

import click
import docker  # type: ignore

from pytezos import ContractInterface
from pytezos import __version__
from pytezos import pytezos
from pytezos.cli.github import create_deployment
from pytezos.cli.github import create_deployment_status
from pytezos.context.mixin import default_network
from pytezos.logging import DEFAULT_LOGGING_CONFIG
from pytezos.logging import logger
from pytezos.michelson.types.base import generate_pydoc
from pytezos.operation.result import OperationResult
from pytezos.rpc.errors import RpcError
from pytezos.sandbox.node import DOCKER_IMAGE
from pytezos.sandbox.node import TEZOS_NODE_PORT
from pytezos.sandbox.node import SandboxedNodeContainer
from pytezos.sandbox.node import get_next_baker_key
from pytezos.sandbox.parameters import KATHMANDU

kernel_js_path = join(dirname(dirname(__file__)), 'assets', 'kernel.js')
kernel_json = {
    "argv": ['pytezos', 'kernel', 'run', "-file", "{connection_file}"],
    "display_name": "Michelson",
    "language": "michelson",
    "codemirror_mode": "michelson",
}

SMARTPY_CLI_IMAGE = 'bakingbad/smartpy-cli'


def make_bcd_link(network, address):
    return f'https://better-call.dev/{network}/{address}'


def get_local_contract_path(path, extension='tz'):
    if path is None:
        files = glob(f'*.{extension}')
        if len(files) != 1:
            raise Exception('No contracts found in working directory, specify --path implicitly')
        path = abspath(files[0])
    if exists(path):
        return path
    return False


def get_contract(path):
    path = get_local_contract_path(path)
    if path:
        contract = ContractInterface.from_file(path)
    else:
        network, address = path.split(':')
        contract = pytezos.using(shell=network).contract(address)
    return contract


def get_docker_client():
    return docker.from_env()


@click.group()
@click.version_option(__version__)
@click.pass_context
def cli(*_args, **_kwargs):
    if not logging.getLogger().hasHandlers():
        logging.config.dictConfig(DEFAULT_LOGGING_CONFIG)


@cli.command(help='Manage contract storage')
@click.option('--action', '-a', type=str, help='One of `schema`, `default`.')
@click.option(
    '--path', '-p', type=str, default=None, help='Path to the .tz file, or the following uri: <network>:<KT-address>'
)
@click.pass_context
def storage(_ctx, action: str, path: Optional[str]) -> None:
    contract = get_contract(path)
    if action == 'schema':
        logger.info(generate_pydoc(type(contract.storage.data), title='storage'))
    elif action == 'default':
        logger.info(pformat(contract.storage.dummy()))
    else:
        raise Exception('Action must be either `schema` or `default`')


@cli.command(help='Manage contract storage')
@click.option('--action', '-a', type=str, default='schema', help='One of `schema`')
@click.option(
    '--path', '-p', type=str, default=None, help='Path to the .tz file, or the following uri: <network>:<KT-address>'
)
@click.pass_context
def parameter(_ctx, action: str, path: Optional[str]) -> None:
    contract = get_contract(path)
    if action == 'schema':
        logger.info(contract.parameter.__doc__)
    else:
        raise Exception('Action must be `schema`')


@cli.command(help='Activate and reveal key from the faucet file')
@click.option('--path', '-p', type=str, help='Path to the .json file downloaded from https://faucet.tzalpha.net/')
@click.option('--network', '-n', type=str, default=default_network, help='Default is florencenet')
@click.pass_context
def activate(_ctx, path: str, network: str) -> None:
    ptz = pytezos.using(key=path, shell=network)
    logger.info(
        'Activating %s in the %s',
        ptz.key.public_key_hash(),
        network,
    )

    if ptz.balance() == 0:
        try:
            opg = ptz.reveal().autofill().sign()
            logger.info('Injecting reveal operation:')
            logger.info(pformat(opg.json_payload()))
            opg.inject(_async=False)
        except RpcError as e:
            logger.critical(pformat(e))
            sys.exit(-1)
        else:
            logger.info('Activation succeeded! Claimed balance: %s ꜩ', ptz.balance())
    else:
        logger.info('Already activated')

    try:
        opg = ptz.reveal().autofill().sign()
        logger.info('Injecting reveal operation:')
        logger.info(pformat(opg.json_payload()))
        opg.inject(_async=False)
    except RpcError as e:
        logger.critical(pformat(e))
        sys.exit(-1)
    else:
        logger.info('Your key %s is now active and revealed', ptz.key.public_key_hash())


@cli.command(help='Deploy contract to the specified network')
@click.option('--path', '-p', type=str, help='Path to the .tz file')
@click.option('--storage', type=str, default=None, help='Storage in JSON format (not Micheline)')
@click.option('--network', '-n', type=str, default=default_network, help='Default is florencenet')
@click.option('--key', type=str, default=None)
@click.option('--github-repo-slug', type=str, default=None)
@click.option('--github-oauth-token', type=str, default=None)
@click.option('--dry-run', type=bool, default=False, help='Set this flag if you just want to see what would happen')
@click.pass_context
def deploy(
    _ctx,
    path: str,
    storage: Optional[str],  # pylint: disable=redefined-outer-name
    network: str,
    key: Optional[str],
    github_repo_slug: Optional[str],
    github_oauth_token: Optional[str],
    dry_run: bool,
):
    ptz = pytezos.using(shell=network, key=key)
    logger.info('Deploying contract using %s in the %s', ptz.key.public_key_hash(), network)

    contract = get_contract(path)
    try:
        opg = ptz.origination(script=contract.script(initial_storage=storage)).autofill().sign()
        logger.info('Injecting origination operation:')
        logger.info(pformat(opg.json_payload()))

        if dry_run:
            logger.info(pformat(opg.preapply()))
            sys.exit(0)
        else:
            opg = opg.inject(_async=False)
    except RpcError as e:
        logger.critical(pformat(e))
        sys.exit(-1)
    else:
        originated_contracts = OperationResult.originated_contracts(opg)
        if len(originated_contracts) != 1:
            raise Exception('Operation group must has exactly one originated contract')
        bcd_link = make_bcd_link(network, originated_contracts[0])
        logger.info('Contract was successfully deployed: %s', bcd_link)

        if github_repo_slug:
            deployment = create_deployment(
                github_repo_slug,
                github_oauth_token,
                environment=network,
            )
            logger.info(pformat(deployment))
            status = create_deployment_status(
                github_repo_slug,
                github_oauth_token,
                deployment_id=deployment['id'],
                state='success',
                environment=network,
                environment_url=bcd_link,
            )
            logger.info(status)


@cli.command(help='Update containerized SmartPy CLI')
@click.option('--tag', '-t', type=str, help='Version or tag to pull', default='latest')
@click.pass_context
def update_smartpy(ctx, tag):
    client = get_docker_client()
    logger.info('Will now pull latest SmartPy image, please stay put.')
    for line in client.api.pull(f'{SMARTPY_CLI_IMAGE}:{tag}', stream=True, decode=True):
        logger.info(line)
    logger.info('Pulled SmartPy CLI image successfully!')


def run_smartpy_container(
    tag: str = 'latest',
    command: str = '',
    files_to_add: Optional[List[str]] = None,
    mounts: Optional[List[docker.types.Mount]] = None,
):
    try:
        client = get_docker_client()
        container = client.containers.create(
            image=f'{SMARTPY_CLI_IMAGE}:{tag}',
            command=command,
            detach=True,
            mounts=mounts or [],
        )
        buffer = io.BytesIO()
        with tarfile.open(fileobj=buffer, mode='w:gz') as archive:
            for filename in files_to_add or []:
                with open(filename, 'rb') as current_file:
                    current_file_data = current_file.read()
                    current_file_buffer = io.BytesIO(initial_bytes=current_file_data)
                    _, short_filename = split(filename)
                    archive.add(filename, arcname=short_filename)
        buffer.seek(0)
        container.put_archive(
            '/root/smartpy-cli/',
            buffer,
        )
        container.start()
        return container
    except docker.errors.ImageNotFound:
        logger.error('SmartPy compiler not found. Please run update-smartpy first.')


@cli.command(help='Run SmartPy CLI command "test"')
@click.option('--script', '-s', type=str, help='Path to script', default='script.py')
@click.option('--output-directory', '-o', type=str, help='Output directory', default='./smartpy-output')
@click.option(
    '--protocol', type=click.Choice(['delphi', 'edo', 'florence', 'proto10']), help='Protocol to use', default='edo'
)
@click.option('--detach', '-d', type=bool, help='Run container in detached mode', default=False)
@click.option('--tag', '-t', type=str, help='Version or tag of SmartPy to use', default='latest')
@click.pass_context
def smartpy_test(
    _ctx,
    script: str,
    output_directory: str,
    detach: bool,
    protocol: str,
    tag: str,
):
    client = get_docker_client()
    path = get_local_contract_path(script, extension='py')
    if path:
        _, script_name = split(path)
        container = run_smartpy_container(
            tag=tag,
            command=f'test /root/smartpy-cli/{script_name} /root/output --protocol {protocol}',
            files_to_add=[path],
            mounts=[
                docker.types.Mount(
                    target='/root/output',
                    source=output_directory,
                    type='bind',
                )
            ],
        )
        if container is None:
            raise Exception('Could not create container. Try running update-smartpy.')
        if not detach:
            for line in container.logs(stream=True):
                print(line.decode('utf-8').rstrip())
    else:
        logger.error('No local script found. Please ensure a valid script is present or specify path.')


@cli.command(help='Run SmartPy CLI command "compile"')
@click.option('--script', '-s', type=str, help='Path to script', default='script.py')
@click.option('--output-directory', '-o', type=str, help='Output directory', default='./smartpy-output')
@click.option('--detach', '-d', type=bool, help='Run container in detached mode', default=False)
@click.option(
    '--protocol', type=click.Choice(['delphi', 'edo', 'florence', 'proto10']), help='Protocol to use', default='edo'
)
@click.option('--tag', '-t', type=str, help='Version or tag of SmartPy to use', default='latest')
@click.pass_context
def smartpy_compile(
    _ctx,
    script: str,
    output_directory: str,
    detach: bool,
    protocol: str,
    tag: str,
):
    client = get_docker_client()
    path = get_local_contract_path(script, extension='py')
    if path:
        _, script_name = split(path)
        container = run_smartpy_container(
            tag=tag,
            command=f'compile /root/smartpy-cli/{script_name} /root/output --protocol {protocol}',
            files_to_add=[path],
            mounts=[
                docker.types.Mount(
                    target='/root/output',
                    source=output_directory,
                    type='bind',
                )
            ],
        )
        if container is None:
            raise Exception('Could not create container. Try running update-smartpy.')
        if not detach:
            for line in container.logs(stream=True):
                print(line.decode('utf-8').rstrip())
    else:
        logger.error('No local script found. Please ensure a valid script is present or specify path.')


@cli.command(help='Run containerized sandbox node')
@click.option('--image', type=str, help='Docker image to use', default=DOCKER_IMAGE)
@click.option('--protocol', type=click.Choice(['kathmandu']), help='Protocol to use', default='kathmandu')
@click.option('--port', '-p', type=int, help='Port to expose', default=TEZOS_NODE_PORT)
@click.option('--interval', '-i', type=float, help='Interval between baked blocks (in seconds)', default=1.0)
@click.option('--blocks', '-b', type=int, help='Number of blocks to bake before exit')
@click.pass_context
def sandbox(
    _ctx,
    image: str,
    protocol: str,
    port: int,
    interval: float,
    blocks: int,
):
    protocol_hash = {
        'kathmandu': KATHMANDU,
    }[protocol]

    node = SandboxedNodeContainer(image=image, port=port)
    with node:
        if not node.wait_for_connection():
            raise TimeoutError('Failed to connect to the sandboxed node')

        logger.info('Activating protocol %s...', protocol_hash)
        node.activate(protocol_hash)

        blocks_baked = 0
        while True:
            try:
                logger.info('Baking block %s...', blocks_baked)
                block_hash = node.bake(key=get_next_baker_key(node.client))
                logger.info('Baked block: %s', block_hash)
                blocks_baked += 1
                if blocks and blocks_baked == blocks:
                    break
                time.sleep(interval)
            except KeyboardInterrupt:
                break


@cli.command(help='Update Ligo compiler (docker pull ligolang/ligo)')
@click.option('--tag', '-t', type=str, help='Version or tag to pull', default='0.13.0')
@click.pass_context
def update_ligo(
    _ctx,
    tag: str,
):
    client = get_docker_client()

    logger.info(f'Pulling ligolang/ligo{(":" + tag) if tag else ""}, please stay put.')
    for line in client.api.pull('ligolang/ligo', tag=tag, stream=True, decode=True):
        logger.info(line)
    logger.info('Pulled Ligo compiler image successfully!')


def run_ligo_container(
    tag: str = '0.13.0',
    command: str = '',
    files_to_add: Optional[List[str]] = None,
):
    try:
        client = get_docker_client()
        container = client.containers.create(
            image=f'ligolang/ligo:{tag}',
            command=command,
            detach=True,
        )
        buffer = io.BytesIO()
        with tarfile.open(fileobj=buffer, mode='w:gz') as archive:
            for filename in files_to_add or []:
                with open(filename, 'rb') as current_file:
                    current_file_data = current_file.read()
                    current_file_buffer = io.BytesIO(initial_bytes=current_file_data)
                    _, short_filename = split(filename)
                    archive.add(filename, arcname=short_filename)
        buffer.seek(0)
        container.put_archive(
            '/root/',
            buffer,
        )
        container.start()
        return container
    except docker.errors.ImageNotFound:
        logger.error('Ligo compiler not found. Please run update-ligo first.')


@cli.command(help='Compile contract using Ligo compiler.')
@click.option('--tag', '-t', type=str, help='Version or tag of Ligo compiler', default='0.13.0')
@click.option('--path', '-p', type=str, help='Path to contract')
@click.option('--entry-point', '-ep', type=str, help='Entrypoint for the invocation')
@click.option('--detach', '-d', type=bool, help='Run container in detached mode', default=False)
@click.pass_context
def ligo_compile_contract(
    _ctx,
    tag: str,
    path: str,
    entry_point: str,
    detach: bool,
):
    path = get_local_contract_path(path, extension='ligo')
    if path:
        _, contract_name = split(path)
        container = run_ligo_container(
            tag=tag,
            command=f'compile-contract {contract_name} "{entry_point}"',
            files_to_add=[path],
        )
        if not detach:
            for line in container.logs(stream=True):
                print(line.decode('utf-8').rstrip())
    else:
        logger.error('No local contract found. Please ensure a valid contract is present or specify path.')


@cli.command(help='Define initial storage using Ligo compiler.')
@click.option('--tag', '-t', type=str, help='Version or tag of Ligo compiler', default='0.13.0')
@click.option('--path', '-p', type=str, help='Path to contract')
@click.option('--entry-point', '-ep', type=str, help='Entrypoint for the storage', default='')
@click.option('--expression', '-ex', type=str, help='Expression for the storage', default='')
@click.option('--detach', '-d', type=bool, help='Run container in detached mode', default=False)
@click.pass_context
def ligo_compile_storage(
    _ctx,
    tag: str,
    path: str,
    entry_point: str,
    expression: str,
    detach: bool,
):
    path = get_local_contract_path(path, extension='ligo')
    if path:
        container = run_ligo_container(
            tag=tag,
            command=f'compile-storage {path} "{entry_point}" "{expression}"',
            files_to_add=[path],
        )
        if not detach:
            for line in container.logs(stream=True):
                print(line.decode('utf-8').rstrip())
    else:
        logger.error('No local contract found. Please ensure a valid contract is present or specify path.')


@cli.command(help='Invoke a contract with a parameter using Ligo compiler.')
@click.option('--tag', '-t', type=str, help='Version or tag of Ligo compiler', default='0.13.0')
@click.option('--path', '-p', type=str, help='Path to contract')
@click.option('--entry-point', '-ep', type=str, help='Entrypoint for the invocation')
@click.option('--expression', '-ex', type=str, help='Expression for the invocation')
@click.option('--detach', '-d', type=bool, help='Run container in detached mode', default=False)
@click.pass_context
def ligo_invoke_contract(
    _ctx,
    tag: str,
    path: str,
    entry_point: str,
    expression: str,
    detach: bool,
):
    path = get_local_contract_path(path, extension='ligo')
    if path:
        container = run_ligo_container(
            tag=tag,
            command=f'compile-parameter {path} "{entry_point}" "{expression}"',
            files_to_add=[path],
        )
        if not detach:
            for line in container.logs(stream=True):
                print(line.decode('utf-8').rstrip())
    else:
        logger.error('No local contract found. Please ensure a valid contract is present or specify path.')


if __name__ == '__main__':
    cli(prog_name='pytezos')



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union
from typing import cast

from pytezos.context.abstract import AbstractContext
from pytezos.michelson.micheline import Micheline
from pytezos.michelson.micheline import MichelsonRuntimeError
from pytezos.michelson.types import OrType
from pytezos.michelson.types.adt import wrap_or
from pytezos.michelson.types.adt import wrap_parameters
from pytezos.michelson.types.base import MichelsonType
from pytezos.michelson.types.base import parse_name
from pytezos.michelson.types.core import Unit


class ParameterSection(Micheline, prim='parameter', args_len=1):
    args: List[Type[MichelsonType]]  # type: ignore
    root_name: str

    def __init__(self, item: MichelsonType):
        super().__init__()
        self.item = item

    def __repr__(self):
        return repr(self.item)

    @staticmethod
    def match(type_expr) -> Type['ParameterSection']:
        try:
            cls = Micheline.match(type_expr)
            if not issubclass(cls, ParameterSection):
                cls = ParameterSection.create_type(args=[cls])
        except Exception as e:
            raise MichelsonRuntimeError('parameter', *e.args) from e
        return cls

    @classmethod
    def create_type(
        cls,
        args: List[Union[Type['Micheline'], Any]],
        annots: Optional[list] = None,
        **kwargs,
    ) -> Type['ParameterSection']:
        assert not annots, 'top level parameter annotations not supported'

        root_type = cast(Type[MichelsonType], args[0])
        if issubclass(root_type, OrType):
            root_name = root_type.field_name  # type: ignore
            if not root_name:
                flat_args = root_type.get_flat_args(entrypoints=True)  # type: ignore
                assert isinstance(flat_args, dict), f'expected a named type, got {flat_args}'
                root_name = 'root' if 'default' in flat_args else 'default'
        else:
            root_name = root_type.field_name or 'default'
        res = type(cls.__name__, (cls,), dict(args=args, root_name=root_name, **kwargs))
        return cast(Type['ParameterSection'], res)

    @classmethod
    def execute(cls, stack, stdout: List[str], context: AbstractContext):
        context.set_parameter_expr(cls.as_micheline_expr())
        stdout.append(f'parameter: updated')

    @classmethod
    def list_entrypoints(cls) -> Dict[str, Type[MichelsonType]]:
        entrypoints = {}
        root_type = cls.args[0]
        if issubclass(root_type, OrType):
            flat_args = root_type.get_flat_args(entrypoints=True)
            assert isinstance(flat_args, dict), f'expected dict of named entrypoints'
            for name, arg in flat_args.items():
                entrypoints[name] = arg.get_anon_type()
        entrypoints[cls.root_name] = root_type
        return entrypoints

    @classmethod
    def from_parameters(cls, parameters: Dict[str, Any]) -> 'ParameterSection':
        if len(parameters) == 0:
            parameters = {'entrypoint': 'default', 'value': {'prim': 'Unit'}}
        assert isinstance(parameters, dict) and parameters.keys() == {
            'entrypoint',
            'value',
        }, f'expected {{entrypoint, value}}, got {parameters}'
        entrypoint = parameters['entrypoint']
        if entrypoint == cls.root_name:
            res = cls.from_micheline_value(parameters['value'])
            return cast(ParameterSection, res)
        else:
            root_type = cls.args[0]
            assert issubclass(root_type, OrType), f'expected `{cls.root_name}`, got `{entrypoint}`'
            _, key_to_path, _ = root_type.get_type_layout(entrypoints=True)
            assert entrypoint in key_to_path, f'unexpected entrypoint `{entrypoint}`'  # type: ignore
            val_expr = wrap_parameters(parameters['value'], key_to_path[entrypoint])  # type: ignore
            item = root_type.from_micheline_value(val_expr)
            return cls(item)

    def to_parameters(self, mode='readable') -> Dict[str, Any]:
        entrypoint, item = self.root_name, self.item
        if isinstance(self.item, OrType):
            flat_values = self.item.get_flat_values(entrypoints=True)
            assert isinstance(flat_values, dict) and len(flat_values) == 1, f'expected named type'
            entrypoint, item = next(iter(flat_values.items()))
        return {
            'entrypoint': entrypoint,
            'value': item.to_micheline_value(mode=mode, lazy_diff=None),
        }

    @classmethod
    def generate_pydoc(cls) -> str:
        definitions = []  # type: ignore
        res = cls.args[0].generate_pydoc(definitions, cls.prim)
        if res != f'${cls.prim}':
            definitions.insert(0, (cls.prim, res))
        return '\n'.join(f'${var}:\n\t{doc}\n' for var, doc in definitions)

    @classmethod
    def from_micheline_value(cls, val_expr) -> 'ParameterSection':
        item = cls.args[0].from_micheline_value(val_expr)
        return cls(item)

    @classmethod
    def from_python_object(cls, py_obj) -> 'ParameterSection':
        if isinstance(py_obj, str):
            entrypoint = py_obj
            py_obj = {entrypoint: Unit}
        else:
            if not isinstance(py_obj, dict) or len(py_obj) != 1:
                raise TypeError(f'expected dict with a single key, got {type(py_obj).__name__} `{py_obj}`')
            entrypoint = next(iter(py_obj))

        if entrypoint == cls.root_name:
            item = cls.args[0].from_python_object(py_obj[entrypoint])
        else:
            if not issubclass(cls.args[0], OrType):
                raise TypeError(f'Unexpected entrypoint `{entrypoint}`: parameter is not of sum type')
            _, key_to_path, _ = cls.args[0].get_type_layout(infer_names=True, entrypoints=True)
            if not key_to_path:
                raise TypeError('sum type has to be named (in the scope of PyTezos)')
            item = cls.args[0].from_python_object(wrap_or(py_obj[entrypoint], key_to_path[entrypoint]))

        return cls(item)

    def to_micheline_value(self, mode='readable', lazy_diff=None):
        return self.item.to_micheline_value(mode=mode, lazy_diff=None)

    def to_python_object(self, try_unpack=False, lazy_diff=None) -> dict:
        py_obj = self.item.to_python_object(
            try_unpack=try_unpack,
            lazy_diff=None,
        )
        if issubclass(self.args[0], OrType):
            return py_obj
        else:
            return {self.root_name: py_obj}

    def merge_lazy_diff(self, lazy_diff: List[dict]) -> 'ParameterSection':
        item = self.item.merge_lazy_diff(lazy_diff)
        return type(self)(item)

    def attach_context(self, context: AbstractContext):
        self.item.attach_context(context, big_map_copy=True)

    def aggregate_lazy_diff(self, mode='readable') -> List[dict]:
        lazy_diff = []  # type: ignore
        self.item.aggregate_lazy_diff(lazy_diff, mode=mode)
        return lazy_diff

    def __getitem__(self, key):
        assert hasattr(self.item, '__getitem__'), f'index access is not implemented for {self.item.prim}'
        return self.item[key]



from pprint import pformat
from typing import List
from typing import Optional
from typing import Tuple

from pytezos.michelson.types.base import MichelsonType


class MichelsonStack:
    def __init__(self, items: Optional[List[MichelsonType]] = None) -> None:
        self.items = items or []
        self.protected = 0

    @classmethod
    def from_items(cls, items: List[MichelsonType]) -> 'MichelsonStack':
        return cls(items)

    def protect(self, count: int) -> None:
        if len(self.items) < count:
            raise Exception(f'got {len(self.items)} items on the stack, want to protect {count}')
        self.protected += count

    def restore(self, count: int) -> None:
        if self.protected < count:
            raise Exception(f'want to restore {count} items, but only {self.protected} are protected')
        self.protected -= count

    def push(self, item: MichelsonType):
        self.items.insert(self.protected, item)

    def peek(self) -> MichelsonType:
        if not self.items:
            raise Exception('stack is empty')
        return self.items[self.protected]

    def pop(self, count: int) -> List[MichelsonType]:
        if len(self.items) - self.protected < count:
            raise Exception(f'got {len(self.items) - self.protected} items on the stack, want to pop {count}')
        return [self.items.pop(self.protected) for _ in range(count)]

    def pop1(self) -> MichelsonType:
        (a,) = self.pop(count=1)
        return a

    def pop2(self) -> Tuple[MichelsonType, MichelsonType]:
        a, b = self.pop(count=2)
        return a, b

    def pop3(self) -> Tuple[MichelsonType, MichelsonType, MichelsonType]:
        a, b, c = self.pop(count=3)
        return a, b, c

    def clear(self) -> None:
        self.items.clear()
        self.protected = 0

    def dump(self, count: int) -> Optional[List[MichelsonType]]:
        if not self.items:
            return None
        count = min(count, len(self.items))
        return self.items[:count]

    def __len__(self) -> int:
        return len(self.items)

    def __repr__(self) -> str:
        return pformat(self.items)



import logging
from decimal import Decimal
from typing import List
from typing import Optional
from typing import Union

from pytezos.block.header import BlockHeader
from pytezos.context.mixin import ContextMixin
from pytezos.contract.call import ContractCall
from pytezos.contract.interface import ContractInterface
from pytezos.crypto.key import Key
from pytezos.jupyter import get_class_docstring
from pytezos.jupyter import is_interactive
from pytezos.logging import logger
from pytezos.operation.content import ContentMixin
from pytezos.operation.group import OperationGroup
from pytezos.rpc import ShellQuery
from pytezos.sandbox.parameters import get_protocol_parameters


class PyTezosClient(ContextMixin, ContentMixin):
    """Entry point for a developer, start your script with:

    .. code-block:: python

        from pytezos import pytezos
    """

    def __repr__(self):
        res = [
            super().__repr__(),
            '\nHelpers',
            get_class_docstring(self.__class__),
        ]
        return '\n'.join(res)

    def operation_group(self, protocol=None, branch=None, contents=None, signature=None) -> OperationGroup:
        """Create new operation group (multiple contents).

        You can leave all fields empty in order to create an empty operation group.

        :param protocol: Leave None for autocomplete, unless you know what you are doing
        :param branch: Leave None for autocomplete
        :param contents: List of operation contents (optional)
        :param signature: Can be set later
        :rtype: OperationGroup
        """
        return OperationGroup(
            context=self._spawn_context(),
            protocol=protocol,
            branch=branch,
            contents=contents,
            signature=signature,
        )

    def operation(self, content: dict) -> OperationGroup:
        """Create an operation group with single content.

        :param content: Operation body (depending on `kind`)
        :rtype: OperationGroup
        """
        return OperationGroup(context=self._spawn_context(), contents=[content])

    def bulk(self, *operations: Union[OperationGroup, ContractCall]) -> OperationGroup:
        """Batch multiple operations and contract calls in a single operation group

        :param operations: a tuple of operations or contract calls
        :rtype: OperationGroup
        """
        contents = []
        reset_fields = {
            'pkh': '',
            'source': '',
            'delegate': '',
            'counter': '0',
            'secret': '',
            'period': '0',
            'public_key': '',
            'fee': '0',
            'gas_limit': '0',
            'storage_limit': '0',
        }
        for opg in operations:
            if isinstance(opg, ContractCall):
                opg = opg.as_transaction()
            else:
                assert isinstance(opg, OperationGroup), f'expected OperationGroup or ContractCall, got {opg}'
            for content in opg.contents:
                contents.append({k: reset_fields.get(k, v) for k, v in content.items()})
        return OperationGroup(context=self._spawn_context(), contents=contents)

    def account(self, account_id=None) -> dict:
        """Shortcut for RPC contract request.

        :param account_id: tz/KT address, leave None to show info about current key
        """
        address = account_id or self.key.public_key_hash()
        return self.shell.contracts[address]()

    def balance(self) -> Decimal:
        """Get account balance.

        :return: amount in tez
        """
        balance_str = self.account()['balance']
        return (Decimal(balance_str) / 10**6).quantize(Decimal('0.000001'))

    def now(self) -> int:
        """Timestamp of the latest block + block time (UTC)."""
        return self.context.get_now()

    def contract(self, address) -> ContractInterface:
        """Get a high-level interface for a given smart contract id.

        :param address: KT address of a smart contract
        :rtype: ContractInterface
        """
        return ContractInterface.from_context(self._spawn_context(address=address))

    def using(
        self,
        shell: Optional[Union[ShellQuery, str]] = None,
        key: Optional[Union[Key, str, dict]] = None,
        mode: Optional[str] = None,
        ipfs_gateway: Optional[str] = None,
    ):
        """Change current RPC endpoint and account (private key).

        :param shell: one of 'mainnet', '***net', or RPC node uri, or instance of :class:`pytezos.rpc.shell.ShellQuery`
        :param key: base58 encoded key, path to the faucet file, faucet file itself, alias from tezos-client, or `Key`
        :param mode: whether to use `readable` or `optimized` encoding for parameters/storage/other
        :returns: A copy of current object with changes applied
        """
        return PyTezosClient(
            context=self._spawn_context(
                shell=shell,
                key=key,
                mode=mode,
                ipfs_gateway=ipfs_gateway,
            )
        )

    @property
    def loglevel(self) -> int:
        return logger.level

    @loglevel.setter
    def loglevel(self, value: Union[str, int]) -> None:
        if is_interactive():
            logging.getLogger().setLevel(value)
        logger.setLevel(value)

    def activate_protocol(self, protocol_hash: str) -> BlockHeader:
        """Initiate user-activated upgrade (sandbox only)

        :param protocol_hash: Protocol hash
        :rtype: BlockHeader
        """
        return BlockHeader.activate_protocol(
            protocol_hash=protocol_hash,
            parameters=get_protocol_parameters(protocol_hash),
            context=self.context,
        )

    def bake_block(self, min_fee: int = 0) -> BlockHeader:
        """Create and inject new block with operations from mempool

        :param min_fee: filter operations by fee (default is 0)
        :rtype: BlockHeader
        """
        return BlockHeader.bake_block(
            min_fee=min_fee,
            context=self.context,
        )

    def sign_message(self, message: str, block: Union[str, int] = 'genesis') -> str:
        """Sign arbitrary message with guarantee that resulting operation won't be used onchain.

        :param message: Message to sign
        :param block: Specify block, defaults to genesis
        :returns: Base58-encoded signature (non-generic)
        """
        return self.key.sign(self.failing_noop(message).message(block=block))

    def check_message(self, message: str, public_key: str, signature: str, block: str = 'genesis') -> None:
        """Check message signature

        :param message: Signed operation
        :param public_key: Signer's public key
        :param signature: Message signature
        :param block: Specify block, defaults to genesis
        """
        pk = Key.from_encoded_key(public_key)
        pk.verify(
            signature=signature,
            message=self.failing_noop(message).message(block=block),
        )

    def wait(
        self,
        *operation_groups: OperationGroup,
        min_confirmations: int = 1,
        num_blocks_wait: int = 5,
        time_between_blocks: Optional[int] = None,
        prev_hash: Optional[str] = None,
        block_timeout: Optional[int] = None,
    ) -> List[dict]:
        """Wait for multiple injected operations get enough confirmations

        :param min_confirmations: number of block injections to wait for before returning
        :param num_blocks_wait: number of blocks to wait for injection
        :param time_between_blocks: override the corresponding parameter from constants
        :param prev_hash: Current block hash (optional). If not set, current head is used.
        :param block_timeout: set block timeout (by default Pytezos will wait for a long time)
        """
        if len(operation_groups) == 0:
            raise ValueError('At least one operation group has to be passed to the args')

        opg_hashes = []
        for opg in operation_groups:
            if opg.opg_hash is None:
                raise ValueError('All operations must have hash assigned')
            opg_hashes.append(opg.opg_hash)

        return self.shell.wait_operations(
            opg_hashes=opg_hashes,
            ttl=num_blocks_wait,
            min_confirmations=min_confirmations,
            current_block_hash=prev_hash,
            time_between_blocks=time_between_blocks,
            block_timeout=block_timeout,
        )

    def sleep(
        self,
        num_blocks: int,
        time_between_blocks: Optional[int] = None,
        block_timeout: Optional[int] = None,
    ) -> List[str]:
        """Sleeps until a certain amount of blocks appended to the chain

        :param num_blocks: number of blocks to wait for
        :param time_between_blocks: override the corresponding parameter from constants
        :param block_timeout: set block timeout (by default Pytezos will wait for a long time)
        """
        block_hash = self.shell.head.hash()
        return list(
            self.shell.wait_blocks(
                current_block_hash=block_hash,
                max_blocks=num_blocks,
                time_between_blocks=time_between_blocks,
                block_timeout=block_timeout,
            )
        )



from pprint import pformat
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from deprecation import deprecated  # type: ignore

from pytezos.context.impl import ExecutionContext
from pytezos.context.mixin import ContextMixin
from pytezos.crypto.encoding import base58_decode
from pytezos.crypto.encoding import base58_encode
from pytezos.crypto.encoding import is_bh
from pytezos.crypto.key import blake2b_32
from pytezos.jupyter import get_class_docstring
from pytezos.logging import logger
from pytezos.michelson.forge import forge_base58
from pytezos.operation import DEFAULT_BURN_RESERVE
from pytezos.operation import DEFAULT_GAS_RESERVE
from pytezos.operation import MAX_OPERATIONS_TTL
from pytezos.operation.content import ContentMixin
from pytezos.operation.fees import calculate_fee
from pytezos.operation.fees import default_fee
from pytezos.operation.fees import default_gas_limit
from pytezos.operation.fees import default_storage_limit
from pytezos.operation.forge import forge_operation_group
from pytezos.operation.result import OperationResult
from pytezos.rpc.errors import RpcError
from pytezos.rpc.kind import validation_passes


class OperationGroup(ContextMixin, ContentMixin):
    """Operation group representation: contents (single or multiple), signature, other fields,
    and also useful helpers for filling with precise fees, signing, forging, and injecting.
    """

    def __init__(
        self,
        context: ExecutionContext,
        contents: Optional[List[Dict[str, Any]]] = None,
        protocol: Optional[str] = None,
        chain_id: Optional[str] = None,
        branch: Optional[str] = None,
        signature: Optional[str] = None,
        opg_hash: Optional[str] = None,
        opg_result: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(context=context)
        self.contents = contents or []
        self.protocol = protocol
        self.chain_id = chain_id
        self.branch = branch
        self.signature = signature
        self.opg_hash = opg_hash
        self.opg_result = opg_result

    def __repr__(self) -> str:
        res = [
            super().__repr__(),
            '\nPayload',
            pformat(self.json_payload()),
            '\nHelpers',
            get_class_docstring(self.__class__),
        ]
        return '\n'.join(res)

    def _spawn(self, **kwargs) -> 'OperationGroup':
        return OperationGroup(
            context=self.context,
            contents=kwargs.get('contents', self.contents.copy()),
            protocol=kwargs.get('protocol', self.protocol),
            chain_id=kwargs.get('chain_id', self.chain_id),
            branch=kwargs.get('branch', self.branch),
            signature=kwargs.get('signature', self.signature),
            opg_hash=kwargs.get('opg_hash', self.opg_hash),
            opg_result=kwargs.get('opg_result', self.opg_result),
        )

    def json_payload(self) -> Dict[str, Any]:
        """Get JSON payload used for the injection."""
        return {
            'protocol': self.protocol,
            'branch': self.branch,
            'contents': self.contents,
            'signature': self.signature,
        }

    def binary_payload(self) -> bytes:
        """Get binary payload used for injection/hash calculation."""
        if not self.signature:
            raise ValueError('Not signed')

        return bytes.fromhex(self.forge()) + forge_base58(self.signature)

    def operation(self, content: Dict[str, Any]) -> 'OperationGroup':
        """Create new operation group with extra content added.

        :param content: Kind-specific operation body
        :rtype: OperationGroup
        """
        return self._spawn(contents=self.contents + [content])

    def fill(
        self,
        counter: Optional[int] = None,
        ttl: Optional[int] = None,
        gas_limit: Optional[int] = None,
        storage_limit: Optional[int] = None,
        minimal_nanotez_per_gas_unit: Optional[int] = None,
        **kwargs,
    ) -> 'OperationGroup':
        """Try to fill all fields left unfilled, use approximate fees
        (not optimal, use `autofill` to simulate operation and get precise values).

        :param counter: Override counter value (for manual handling)
        :param ttl: Number of blocks to wait in the mempool before removal (default is 5 for public network, 60 for sandbox)
        :param gas_limit: Override gas_limit value (for manual handling)
        :param storage_limit: Override storage_limit value (for manual handling)
        :param minimal_nanotez_per_gas_unit: Override minimal_nanotez_per_gas_unit constant (for manual handling)
        :rtype: OperationGroup
        """
        if kwargs.get('branch_offset') is not None:
            logger.warning('`branch_offset` argument is deprecated, use `ttl` instead')
            ttl = MAX_OPERATIONS_TTL - kwargs['branch_offset']

        if ttl is None:
            ttl = self.context.get_operations_ttl()
        if not 0 < ttl <= MAX_OPERATIONS_TTL:
            raise Exception(f'`ttl` has to be in range (0, {MAX_OPERATIONS_TTL}]')

        chain_id = self.chain_id or self.context.get_chain_id()
        protocol = self.protocol or self.context.get_protocol()
        branch = self.branch or self.shell.blocks[f'head~{MAX_OPERATIONS_TTL - ttl}'].hash()
        source = self.key.public_key_hash()
        constants = self.shell.head.context.constants()

        if counter is not None:
            self.context.set_counter(counter - 1)  # which is supposedly the current state (head)

        if gas_limit is None:
            hard_gas_limit_per_content = int(constants['hard_gas_limit_per_operation']) // len(self.contents)
        else:
            hard_gas_limit_per_content = gas_limit // len(self.contents)

        if storage_limit is None:
            hard_storage_limit_per_content = int(constants['hard_storage_limit_per_operation']) // len(self.contents)
        else:
            hard_storage_limit_per_content = storage_limit // len(self.contents)

        replace_map = {
            'pkh': source,
            'source': source,
            'delegate': source,  # self registration
            'counter': lambda i, x: str(self.context.get_counter()),
            'secret': lambda i, x: self.key.activation_code,
            'period': lambda i, x: str(self.shell.head.voting_period()),
            'public_key': lambda i, x: self.key.public_key(),
            'gas_limit': lambda i, x: str(
                min(
                    hard_gas_limit_per_content,
                    gas_limit if gas_limit is not None else default_gas_limit(x, constants),
                )
            ),
            'storage_limit': lambda i, x: str(
                min(
                    hard_storage_limit_per_content,
                    storage_limit if storage_limit is not None else default_storage_limit(x, constants),
                )
            ),
            'fee': lambda i, x: str(default_fee(x, gas_limit, minimal_nanotez_per_gas_unit) if i == 0 else 0),
        }

        def fill_content(idx, content):
            content = content.copy()
            for k, v in replace_map.items():
                if content.get(k) in ['', '0']:
                    content[k] = v(idx, content) if callable(v) else v
            return content

        return self._spawn(
            contents=[fill_content(idx=i, content=x) for i, x in enumerate(self.contents)],
            protocol=protocol,
            chain_id=chain_id,
            branch=branch,
        )

    def run(self, block_id: str = 'head'):
        """Simulate operation without signature checks.

        :param block_id: Specify a level at which this operation should be applied (default is head)
        :returns: RPC response from `run_operation`
        """
        return self.shell.blocks[block_id].helpers.scripts.run_operation.post(
            {
                'operation': {
                    'branch': self.branch,
                    'contents': self.contents,
                    'signature': base58_encode(b'0' * 64, b'sig').decode(),
                },
                'chain_id': self.chain_id,
            }
        )

    def forge(self, validate=False) -> str:
        """Convert json representation of the operation group into bytes.

        :param validate: Forge remotely also and compare results, default is False
        :returns: Hex string
        """
        payload = {
            'branch': self.branch,
            'contents': self.contents,
        }
        local_data = forge_operation_group(payload).hex()

        if validate:
            remote_data = self.shell.blocks[self.branch].helpers.forge.operations.post(payload)
            if local_data != remote_data:
                raise ValueError(f'Local forge result differs from remote one:\n\n{local_data}\n\n{remote_data}')

        return local_data

    def message(self, block: Union[str, int] = 'genesis') -> bytes:
        """Get payload for the failing noop operation

        :param block: Specify operation branch (default is genesis)
        :returns: Message bytes
        """
        if len(self.contents) != 1 or self.contents[0]['kind'] != 'failing_noop':
            raise NotImplementedError('Use for signing messages only')

        branch = block if is_bh(str(block)) else self.shell.blocks[block].hash()
        return b'\x03' + bytes.fromhex(self._spawn(branch=branch).forge())

    def autofill(
        self,
        gas_reserve: int = DEFAULT_GAS_RESERVE,
        burn_reserve: int = DEFAULT_BURN_RESERVE,
        counter: Optional[int] = None,
        ttl: Optional[int] = None,
        fee: Optional[int] = None,
        gas_limit: Optional[int] = None,
        storage_limit: Optional[int] = None,
        **kwargs,
    ) -> 'OperationGroup':
        """Fill the gaps and then simulate the operation in order to calculate fee, gas/storage limits.

        :param gas_reserve: Add a safe reserve for dynamically calculated gas limit (default is 100).
        :param burn_reserve: Add a safe reserve for dynamically calculated storage limit (default is 100).
        :param counter: Override counter value (for manual handling)
        :param ttl: Number of blocks to wait in the mempool before removal (default is 5 for public network, 60 for sandbox)
        :param fee: Explicitly set fee for operation. If not set fee will be calculated depending on results of operation dry-run.
        :param gas_limit: Explicitly set gas limit for operation. If not set gas limit will be calculated depending on results of
            operation dry-run. In case of batch will be evenly split between operations.
        :param storage_limit: Explicitly set storage limit for operation. If not set storage limit will be calculated depending on
            results of operation dry-run. In case of batch will be evenly split between operations.
        :rtype: OperationGroup
        """
        if kwargs.get('branch_offset') is not None:
            logger.warning('`branch_offset` argument is deprecated, use `ttl` instead')
            ttl = MAX_OPERATIONS_TTL - kwargs['branch_offset']

        opg = self.fill(counter=counter, ttl=ttl)
        opg_with_metadata = opg.run()
        if not OperationResult.is_applied(opg_with_metadata):
            raise RpcError.from_errors(OperationResult.errors(opg_with_metadata))

        fee_acc = 0
        extra_size = 32 + 64  # size of serialized branch and signature + safe reserve
        num_contents = len(opg_with_metadata['contents'])
        counter_offset = self.context.get_counter_offset()
        opg.contents.clear()

        for content in opg_with_metadata['contents']:
            if validation_passes[content['kind']] == 3:
                if gas_limit is not None:
                    gas_limit_new = gas_limit // num_contents
                else:
                    gas_limit_new = OperationResult.consumed_gas(content)
                    if content['kind'] in ['origination', 'transaction']:
                        gas_limit_new += gas_reserve

                if storage_limit is not None:
                    storage_limit_new = storage_limit // num_contents
                else:
                    paid_storage_size_diff = OperationResult.paid_storage_size_diff(content)
                    burned = OperationResult.burned(content)
                    storage_limit_new = paid_storage_size_diff + burned
                    if content['kind'] in ['origination', 'transaction']:
                        storage_limit_new += burn_reserve

                current_counter = int(content['counter'])
                content.update(
                    counter=str(current_counter + counter_offset),
                    gas_limit=str(gas_limit_new),
                    storage_limit=str(storage_limit_new),
                    fee='0',
                )
                fee_acc += calculate_fee(content, gas_limit_new, extra_size=1 + extra_size // num_contents)

            content.pop('metadata')
            logger.debug("autofilled transaction content: %s" % content)
            opg.contents.append(content)

        if fee or fee_acc:
            opg.contents[0]['fee'] = str(fee if fee is not None else fee_acc)

        return opg

    def sign(self) -> 'OperationGroup':
        """Sign the operation group with the key specified by `using`.

        :rtype: OperationGroup
        """
        validation_pass = validation_passes[self.contents[0]['kind']]
        if any(map(lambda x: validation_passes[x['kind']] != validation_pass, self.contents)):
            raise ValueError('Mixed validation passes')

        if validation_pass == 0:
            if self.chain_id is None:
                raise ValueError('Chain ID is undefined, run .fill first')
            watermark = b'\x02' + base58_decode(self.chain_id.encode())
        else:
            watermark = b'\x03'

        message = watermark + bytes.fromhex(self.forge())
        signature = self.key.sign(message=message, generic=True)

        return self._spawn(signature=signature)

    def hash(self) -> str:
        """Calculate the Base58 encoded operation group hash."""
        hash_digest = blake2b_32(self.binary_payload()).digest()
        return base58_encode(hash_digest, b'o').decode()

    def run_operation(self, block_id: str = 'head'):
        """Simulate operation without signature checks.

        :param block_id: Specify a level at which this operation should be applied (default is head)
        :returns: RPC response from `run_operation`
        """
        return self.run(block_id)

    @deprecated(deprecated_in='3.1.0', removed_in='4.0.0', details='use `run_operation()` instead')
    def preapply(self):
        """Preapply signed operation group.

        :returns: RPC response from `preapply`
        """
        if not self.signature:
            raise ValueError('Not signed')

        return self.run_operation()

    def send(
        self,
        gas_reserve: int = DEFAULT_GAS_RESERVE,
        burn_reserve: int = DEFAULT_BURN_RESERVE,
        min_confirmations: int = 0,
        ttl: Optional[int] = None,
    ) -> 'OperationGroup':
        """

        :param gas_reserve: Add a safe reserve for dynamically calculated gas limit (default is 100).
        :param burn_reserve: Add a safe reserve for dynamically calculated storage limit (default is 100).
        :param min_confirmations: number of block injections to wait for before returning (default is 0, i.e. async mode)
        :param ttl: Number of blocks to wait in the mempool before removal (default is 5 for public network, 60 for sandbox)
        :return: OperationGroup with hash filled
        """
        if ttl is None:
            ttl = self.context.get_operations_ttl()

        opg = self.autofill(gas_reserve=gas_reserve, burn_reserve=burn_reserve, ttl=ttl).sign()
        res = opg.inject(min_confirmations=min_confirmations, num_blocks_wait=ttl)
        return opg._spawn(opg_hash=res['hash'], opg_result=res)

    def send_async(
        self,
        ttl: int,
        counter: int,
        gas_limit: int,
        storage_limit: int,
        minimal_nanotez_per_gas_unit: Optional[int] = None,
    ) -> 'OperationGroup':
        """
        Send operation without simulation or pre-validation

        :param ttl: Number of blocks to wait in the mempool before removal (default is 5 for public network, 60 for sandbox)
        :param counter: Set counter value
        :param gas_limit: Set gas_limit value
        :param storage_limit: Set storage_limit value
        :param minimal_nanotez_per_gas_unit: Override minimal_nanotez_per_gas_unit constant
        :rtype: OperationGroup
        """
        opg = self.fill(
            counter=counter,
            ttl=ttl,
            gas_limit=gas_limit,
            storage_limit=storage_limit,
            minimal_nanotez_per_gas_unit=minimal_nanotez_per_gas_unit,
        ).sign()
        res = opg.inject(prevalidate=False)
        return opg._spawn(opg_hash=res['hash'])

    def inject(
        self,
        check_result: bool = True,
        num_blocks_wait: int = 5,
        time_between_blocks: Optional[int] = None,
        block_timeout: Optional[int] = None,
        min_confirmations: int = 0,
        prevalidate: bool = True,
        **kwargs,
    ):
        """Inject the signed operation group.

        :param check_result: raise RpcError in case operation is applied but has runtime errors
        :param num_blocks_wait: number of blocks to wait for injection
        :param time_between_blocks: override the corresponding parameter from constants
        :param block_timeout: set block timeout (by default Pytezos will wait for a long time)
        :param min_confirmations: number of block injections to wait for before returning
        :param prevalidate: ask node to pre-validate the operation before the injection (True by default)
        :returns: operation group with metadata (raw RPC response)
        """
        self.context.reset()  # reset counter

        opg_hash = self.shell.injection.operation.post(
            operation=self.binary_payload(),
            _async=not prevalidate,
        )

        if min_confirmations == 0:
            return {
                'chain_id': self.chain_id,
                'hash': opg_hash,
                **self.json_payload(),
            }

        operations = self.shell.wait_operations(
            opg_hashes=[opg_hash],
            ttl=num_blocks_wait,
            min_confirmations=min_confirmations,
            time_between_blocks=time_between_blocks,
            block_timeout=block_timeout,
        )

        assert len(operations) == 1
        if check_result:
            if not OperationResult.is_applied(operations[0]):
                raise RpcError.from_errors(OperationResult.errors(operations[0]))

        return operations[0]

    @deprecated(deprecated_in='3.1.0', removed_in='4.0.0', details='use `run_operation()` instead')
    def result(self) -> List[OperationResult]:
        """Parse the preapply result.

        :rtype: List[OperationResult]
        """
        return OperationResult.from_operation_group(self.preapply())
